import click
from click.testing import CliRunner
from omi.cli import translate
import json

def test_cli_translation():
    @click.command()
    @click.argument('name')
    def hello(name):
        click.echo('Hello %s!' % name)

    runner = CliRunner()
    result = runner.invoke(translate, ["-f", "oep-v1.3", "-t", "oep-v1.4", "tests/data/metadata_v13.json"])
    with open("tests/data/metadata_v13_converted.json") as expected_file:
        expected = json.loads(expected_file.read())
        assert result.exit_code == 0
        assert json.loads(result.output) == expected



import json
from collections import OrderedDict

from omi.dialects.oep.compiler import JSONCompiler
from omi.dialects.oep.parser import JSONParser_1_3

from .test_compiler import assert_equal


def test_translation_1_3_to_1_4():
    parser = JSONParser_1_3()
    compiler = JSONCompiler()
    with open("tests/data/metadata_v13_minimal.json", "r") as _input_file:
        input_string = _input_file.read()
        # Step 1: Parse JSON to internal structure
        internal_metadata = parser.parse_from_string(input_string)
        # Step 2: Translate to version 1_4
        result_json = compiler.visit(internal_metadata)

        expected_json = OrderedDict(json.loads('''{
    "metaMetadata": {
        "metadataVersion": "OEP-1.4.0",
        "metadataLicense": {
            "name": "CC0-1.0",
            "title": "Creative Commons Zero v1.0 Universal",
            "path": "https://creativecommons.org/publicdomain/zero/1.0/"}}}'''))

        assert_equal(expected_json, result_json)



import unittest
from omi.dialects.oep.parser import JSONParser_1_3
from omi.dialects.oep.parser import JSONParser_1_4
from omi.dialects.oep.parser import JSONParser_1_5

from ..base.parser import _test_generic_parsing
from ..internal_structures import metadata_v_1_3
from ..internal_structures import metadata_v_1_3_minimal
from ..internal_structures import metadata_v_1_4
from ..internal_structures import metadata_v_1_4_minimal
from ..internal_structures import metadata_v_1_5
from ..internal_structures import metadata_v_1_5_minimal
from omi.dialects.base.parser import ParserException


class ParserTest(unittest.TestCase):
    def test_parser_v1_3(self):
        parser = JSONParser_1_3()
        _input_file = "tests/data/metadata_v13.json"
        expected_result = metadata_v_1_3
        _test_generic_parsing(parser, _input_file, expected_result, nulls=["none", ""])

    def test_parser_v1_4(self):
        parser = JSONParser_1_4()
        _input_file = "tests/data/metadata_v14.json"
        expected_result = metadata_v_1_4
        _test_generic_parsing(parser, _input_file, expected_result)

    def test_parser_v1_4_Resource(self):
        parser = JSONParser_1_4()
        _input_file = "tests/data/metadata_v14_withoutresource.json"
        expected_result = metadata_v_1_4
        with self.assertRaises(ParserException):
            _test_generic_parsing(parser, _input_file, expected_result)

    def test_parser_v1_3_minimal(self):
        parser = JSONParser_1_3()
        _input_file = "tests/data/metadata_v13_minimal.json"
        expected_result = metadata_v_1_3_minimal
        _test_generic_parsing(parser, _input_file, expected_result)

    def test_parser_v1_4_minimal(self):
        parser = JSONParser_1_4()
        _input_file = "tests/data/metadata_v14_minimal.json"
        expected_result = metadata_v_1_4_minimal
        _test_generic_parsing(parser, _input_file, expected_result)

    def test_parser_v1_5(self):
        parser = JSONParser_1_5()
        _input_file = "tests/data/metadata_v15.json"
        expected_result = metadata_v_1_5
        _test_generic_parsing(parser, _input_file, expected_result)



from rdflib import Namespace

DCAT = Namespace("http://www.w3.org/ns/dcat#")
OEO = Namespace("http://openenergy-platform.org/ontology/v0.0.1/oeo/")
DCATDE = Namespace("http://dcat-ap.de/def/dcatde/")
SCHEMA = Namespace("http://schema.org/")
SKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
SPDX = Namespace("http://spdx.org/rdf/terms#")
ADMS = Namespace("http://www.w3.org/ns/adms#")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

from dateutil.parser import parse as parse_date

from omi import structure
from omi.dialects.base.parser import Parser
from omi.dialects.base.parser import ParserException
from omi.oem_structures import oem_v15


def parse_date_or_none(x, *args, **kwargs):
    if x is None:
        return None
    else:
        return parse_date(x, *args, **kwargs)


class JSONParser(Parser):
    def load_string(self, string: str, *args, **kwargs):
        return json.loads(string)

    def is_valid(self, inp: str):
        """Checks the validity of a JSON string

        Parameters
        ----------
        inp: str
            The JSON string to be checked.

        Returns
        -------
        bool
            True if valid JSON, False otherwise.
        """

        try:
            json.loads(inp)
        except ValueError:
            return False
        return True


class JSONParser_1_3(JSONParser):
    def is_valid(self, inp: str):
        if not super(self, JSONParser_1_3).is_valid(inp):
            return False
        try:
            self.assert_1_3_metastring(inp)
        except:
            return False
        else:
            return True

    def parse(self, json_old, *args, **kwargs):
        # context section
        context = None

        # filling the spatial section
        if "spatial" in json_old:
            old_spatial = json_old.get("spatial")
            spatial = structure.Spatial(
                extent=old_spatial.get("extent"),
                resolution=old_spatial.get("resolution"),
            )
        else:
            spatial = None

        # filling the temporal section
        old_temporal = json_old.get("temporal")
        if old_temporal is None:
            temporal = None
        else:
            temporal = structure.Temporal(
                reference_date=parse_date_or_none(old_temporal.get("reference_date"))
            )

        # filling the source section
        # For future reference: There is an important semantic difference between `source = None` and `sources = []`
        # The former means that there is no information regarding sources the latter means that there are no sources.
        # This is holds for all lists around here
        old_sources = json_old.get("sources")
        if old_sources is None:
            sources = None
        else:
            sources = [
                structure.Source(
                    title=old_source.get("name"),
                    description=old_source.get("description"),
                    path=old_source.get("url"),
                    licenses=[
                        structure.TermsOfUse(attribution=old_source.get("copyright"))
                    ],
                )
                for old_source in old_sources
            ]

        # filling the license section
        old_license = json_old.get("license")
        if old_license is None:
            licenses = None  # not []! (see sources)
        else:
            licenses = [
                structure.TermsOfUse(
                    lic=structure.License(
                        identifier=old_license.get("id"),
                        name=old_license.get("name"),
                        path=old_license.get("url"),
                    ),
                    instruction=old_license.get("instruction"),
                    attribution=old_license.get("copyright"),
                )
            ]

        # filling the contributers section
        old_contributors = json_old.get("contributors")
        if old_contributors is None:
            contributions = None
        else:
            contributions = [
                structure.Contribution(
                    contributor=structure.Person(
                        name=old_contributor.get("name"),
                        email=old_contributor.get("email"),
                    ),
                    date=parse_date_or_none(old_contributor.get("date")),
                    comment=old_contributor.get("comment"),
                )
                for old_contributor in old_contributors
            ]

        # extending with script-user information

        old_resources = json_old.get("resources")
        if old_resources is None:
            resources = None
        else:
            resources = []
            if len(old_resources) == 0:
                raise ParserException("Resource field doesn't have any child entity")
            for resource in old_resources:
                old_fields = resource.get("fields")
                if old_fields is None:
                    fields = None
                else:
                    fields = [
                        structure.Field(
                            name=field.get("name"),
                            description=field.get("description"),
                            unit=field.get("unit"),
                        )
                        for field in old_fields
                    ]
                schema = structure.Schema(fields=fields)
                resources.append(
                    structure.Resource(
                        name=resource.get("name"),
                        resource_format="PostgreSQL",
                        schema=schema,
                    )
                )

        review = None

        comment = None

        metadata = structure.OEPMetadata(
            title=json_old.get("title"),
            description=json_old.get("description"),
            languages=json_old.get("language"),
            identifier=None,
            context=context,
            spatial=spatial,
            temporal=temporal,
            sources=sources,
            terms_of_use=licenses,
            contributions=contributions,
            resources=resources,
            review=review,
            comment=comment,
        )
        return metadata


class JSONParser_1_4(JSONParser):
    def is_valid(self, inp: str):
        if not super(self, JSONParser_1_4).is_valid(inp):
            return False
        try:
            self.assert_1_3_metastring(inp)
        except:
            return False
        else:
            return True

    def parse_term_of_use(self, old_license: dict):
        return structure.TermsOfUse(
            lic=structure.License(
                identifier=old_license.get("name"),
                name=old_license.get("title"),
                path=old_license.get("path"),
            ),
            instruction=old_license.get("instruction"),
            attribution=old_license.get("attribution"),
        )

    def parse(self, json_old: dict, *args, **kwargs):
        # context section
        if "id" not in json_old:
            raise ParserException("metadata string does not contain an id")

        inp_context = json_old.get("context")
        if inp_context is None:
            context = None
        else:

            funding_agency = None
            if "fundingAgency" in inp_context:
                funding_agency = structure.Agency(
                    name=inp_context.get("fundingAgency"),
                    logo=inp_context.get("fundingAgencyLogo"),
                )

            context = structure.Context(
                homepage=inp_context.get("homepage"),
                documentation=inp_context.get("documentation"),
                source_code=inp_context.get("sourceCode"),
                contact=inp_context.get("contact"),
                grant_number=inp_context.get("grantNo"),
                funding_agency=funding_agency,
                publisher=structure.Agency(logo=inp_context.get("publisherLogo"))
                if "publisherLogo" in inp_context
                else None,
            )

        # filling the spatial section
        old_spatial = json_old.get("spatial")
        if old_spatial is None:
            spatial = None
        else:
            spatial = structure.Spatial(
                location=old_spatial.get("location"),
                extent=old_spatial.get("extent"),
                resolution=old_spatial.get("resolution"),
            )

        # filling the temporal section
        inp_temporal = json_old.get("temporal")
        if inp_temporal is None:
            temporal = None
        else:
            inp_timeseries = inp_temporal.get("timeseries")
            timeseries = {}
            if inp_timeseries is not None:
                timeseries = dict(
                    start=parse_date_or_none(inp_timeseries.get("start")),
                    end=parse_date_or_none(inp_timeseries.get("end")),
                    resolution=inp_timeseries.get("resolution"),
                    ts_orientation=structure.TimestampOrientation.create(
                        inp_timeseries.get("alignment")
                    )
                    if "alignment" in inp_timeseries
                    and inp_timeseries["alignment"] is not None
                    else None,
                    aggregation=inp_timeseries.get("aggregationType"),
                )
            temporal = structure.Temporal(
                reference_date=parse_date_or_none(inp_temporal.get("referenceDate")),
                **timeseries
            )

        # filling the source section
        old_sources = json_old.get("sources")
        if old_sources is None:
            sources = None
        else:
            sources = [
                structure.Source(
                    title=old_source.get("title"),
                    description=old_source.get("description"),
                    path=old_source.get("path"),
                    licenses=[
                        self.parse_term_of_use(l)
                        for l in old_source.get("licenses", [])
                    ],
                )
                for old_source in old_sources
            ]

        # filling the license section
        old_licenses = json_old.get("licenses")
        if old_licenses is None:
            licenses = None
        else:
            licenses = [
                self.parse_term_of_use(old_license) for old_license in old_licenses
            ]

        # filling the contributers section
        old_contributors = json_old.get("contributors")
        if old_contributors is None:
            contributors = None
        else:
            contributors = [
                structure.Contribution(
                    contributor=structure.Person(
                        name=old_contributor.get("title"),
                        email=old_contributor.get("email"),
                    ),
                    date=parse_date_or_none(old_contributor.get("date")),
                    obj=old_contributor.get("object"),
                    comment=old_contributor.get("comment"),
                )
                for old_contributor in old_contributors
            ]

        # extending with script-user information
        old_resources = json_old.get("resources")
        if old_resources is None:
            resources = None
        # Code added to raise exception when resource is empty
        else:
            if len(old_resources) == 0:
                raise ParserException("Resource field doesn't have any child entity")
            resources = []
            for resource in old_resources:
                old_schema = resource.get("schema")
                if old_schema is None:
                    schema = None
                else:
                    old_fields = old_schema.get("fields")
                    if old_fields is None:
                        fields = None
                    else:
                        fields = [
                            structure.Field(
                                name=field.get("name"),
                                description=field.get("description"),
                                field_type=field.get("type"),
                                unit=field.get("unit"),
                            )
                            for field in old_fields
                        ]
                    field_dict = {field.name: field for field in fields or []}
                    old_foreign_keys = old_schema.get("foreignKeys", [])
                    foreign_keys = []
                    for fk in old_foreign_keys:
                        old_reference = fk.get("reference")
                        if old_reference is None:
                            raise ParserException("Foreign key without reference:", fk)
                        source_fields = [
                            field_dict[field_name]
                            for field_name in fk.get("fields", [])
                        ]
                        old_referenced_fields = old_reference.get("fields")
                        if old_referenced_fields is None:
                            referenced_fields = None
                        else:
                            referenced_fields = [
                                structure.Field(name=fk_field)
                                for fk_field in old_referenced_fields
                            ]
                        referenced_resource = structure.Resource(
                            name=old_reference.get("resource"),
                            schema=structure.Schema(fields=referenced_fields),
                        )
                        for rf in referenced_fields:
                            rf.resource = referenced_resource
                        references = [
                            structure.Reference(s, t)
                            for s, t in zip(source_fields, referenced_fields)
                        ]
                        foreign_keys.append(structure.ForeignKey(references=references))
                    schema = structure.Schema(
                        fields=fields,
                        primary_key=resource["schema"].get("primaryKey"),
                        foreign_keys=foreign_keys,
                    )
                old_dialect = resource.get("dialect")
                if old_dialect is None:
                    dialect = None
                else:
                    dialect = structure.Dialect(
                        delimiter=resource["dialect"].get("delimiter"),
                        decimal_separator=resource["dialect"].get("decimalSeparator"),
                    )
                resources.append(
                    structure.Resource(
                        profile=resource.get("profile"),
                        name=resource.get("name"),
                        path=resource.get("path"),
                        resource_format=resource.get("format"),
                        encoding=resource.get("encoding"),
                        schema=schema,
                        dialect=dialect,
                    )
                )

        inp_review = json_old.get("review")
        if inp_review is None:
            review = None
        else:
            review = structure.Review(
                path=inp_review.get("path"), badge=inp_review.get("badge")
            )

        inp_comment = json_old.get("_comment")
        if inp_comment is None:
            comment = None
        else:
            comment = structure.MetaComment(
                metadata_info=inp_comment.get("metadata"),
                dates=inp_comment.get("dates"),
                units=inp_comment.get("units"),
                languages=inp_comment.get("languages"),
                licenses=inp_comment.get("licenses"),
                review=inp_comment.get("review"),
                none=inp_comment.get("null"),
            )

        metadata = structure.OEPMetadata(
            name=json_old.get("name"),
            title=json_old.get("title"),
            identifier=json_old["id"],
            description=json_old.get("description"),
            languages=json_old.get("language"),
            keywords=json_old.get("keywords"),
            publication_date=parse_date_or_none(json_old.get("publicationDate")),
            context=context,
            spatial=spatial,
            temporal=temporal,
            sources=sources,
            terms_of_use=licenses,
            contributions=contributors,
            resources=resources,
            review=review,
            comment=comment,
        )
        return metadata

    def assert_1_3_metastring(self, json_string: str):
        """Checks string conformity to OEP Metadata Standard Version 1.3

        Parameters
        ----------
        json_string: str
            The JSON string to be checked.

        Returns
        -------
        bool
            True if valid, Raises Exception otherwise.
        """

        keys = [
            "title",
            "description",
            "language",
            "spatial",
            "temporal",
            "sources",
            "license",
            "contributions",
            "resources",
            "metadata_version",
        ]
        subkeys_spatial = ["location", "extent", "resolution"]
        subkeys_temporal = ["reference_date", "start", "end", "resolution"]
        subkeys_license = ["id", "name", "version", "url", "instruction", "copyright"]
        object_subkeys = {
            "spatial": subkeys_spatial,
            "temporal": subkeys_temporal,
            "license": subkeys_license,
        }
        subkeys_sources = [
            "name",
            "description",
            "url",
            "license",
            "copyright",
        ]  # in list of objects
        subkeys_contributors = [
            "name",
            "email",
            "date",
            "comment",
        ]  # in list of objects
        subkeys_resources = ["name", "format", "fields"]  # in list of objects
        list_subkeys = {
            "sources": subkeys_sources,
            "contributions": subkeys_contributors,
            "resources": subkeys_resources,
        }
        subkeys_resources_fields = ["name", "description", "unit"]  # in list of objects

        json_dict = json.loads(json_string)
        try:
            # check if all top level keys are present
            for i in keys:
                if not i in json_dict.keys():
                    raise Exception(
                        'The String did not contain the key "{0}"'.format(i)
                    )
            # check for all keys in second level objects
            for key in object_subkeys:
                for subkey in object_subkeys[key]:
                    if not subkey in json_dict[key]:
                        raise Exception(
                            'The "{0}" object did not contain a "{1}" key'.format(
                                key, subkey
                            )
                        )
            # check for all objects in lists if they contain all required keys
            for key in list_subkeys:
                for list_element in json_dict[key]:
                    for subkey in list_subkeys[key]:
                        if not subkey in list_element:
                            raise Exception(
                                'An object in "{0}" is missing a "{1}" key'.format(
                                    key, subkey
                                )
                            )
        except Exception as error:
            print(
                "The input String does not conform to metadatastring version 1.3 standard"
            )
            print(error)

    # TODO make function check all subkeys as well
    def has_rogue_keys(self, json_string):
        """Checks all keys if they are part of the metadata specification. Gives warnings if not.

        Parameters
        ----------
        json_string: str
            The JSON string to be checked.

        Returns
        -------
        """

        json_dict = json.loads(json_string)
        allowed_keys = [
            "title",
            "description",
            "language",
            "spatial",
            "temporal",
            "sources",
            "license",
            "contributions",
            "resources",
            "metadata_version",
        ]
        for j in json_dict.keys():
            if not j in allowed_keys:
                print('Warning: "{0}" is not among the allowed keys'.format(j))

    def get_table_name(self, metadata_file):
        """Provides the tablename information from the metadata_file

        Parameters
        ----------
        metadata_file: str
            The metadata script from where the tablename is extracted from.

        Returns
        -------
        tablename: str
            returns the tablename.
        """

        raise NotImplementedError


# TODO: Update parser below


class JSONParser_1_5(JSONParser):
    def is_valid(self, inp: str):
        if not super(self, JSONParser_1_5).is_valid(inp):
            return False
        try:
            self.assert_1_5_metastring(inp)
        except:
            return False
        else:
            return True

    def parse_from_string(
        self,
        string: str,
        load_args=None,
        parse_args=None,
        load_kwargs=None,
        parse_kwargs=None,
    ) -> oem_v15.OEPMetadata:
        """
        Parse a string into :class:`~omi.oem_structures.oem_v15.OEPMetadata`

        Parameters
        ----------
        string

        Returns
        -------

        """
        return self.parse(
            self.load_string(string, *(load_args or []), **(load_kwargs or {})),
            *(parse_args or []),
            **(parse_kwargs or {})
        )

    def parse_term_of_use(self, old_license: dict):
        return oem_v15.TermsOfUse(
            lic=oem_v15.License(
                identifier=old_license.get("name"),
                name=old_license.get("title"),
                path=old_license.get("path"),
            ),
            instruction=old_license.get("instruction"),
            attribution=old_license.get("attribution"),
        )

    def parse_timeseries(self, old_timeseries: dict):
        pass

    def parse(self, json_old: dict, *args, **kwargs):
        """_summary_

        Args:
            json_old (dict): _description_

        Raises:
            ParserException: _description_
            ParserException: _description_
            ParserException: _description_

        Returns:
            _type_: _description_
        """

        if "id" not in json_old:
            raise ParserException("metadata string does not contain an id")

        # filling the subject section
        old_subjects = json_old.get("subject")
        if old_subjects is None:
            subject = None
        else:
            subject = [
                oem_v15.Subject(
                    name=old_subject.get("name"), path=old_subject.get("path")
                )
                for old_subject in old_subjects
            ]

        # context section
        inp_context = json_old.get("context")
        if inp_context is None:
            context = None
        else:

            funding_agency = None
            if "fundingAgency" in inp_context:
                funding_agency = oem_v15.Agency(
                    name=inp_context.get("fundingAgency"),
                    logo=inp_context.get("fundingAgencyLogo"),
                )

            context = oem_v15.Context(
                homepage=inp_context.get("homepage"),
                documentation=inp_context.get("documentation"),
                source_code=inp_context.get("sourceCode"),
                contact=inp_context.get("contact"),
                grant_number=inp_context.get("grantNo"),
                funding_agency=funding_agency,
                publisher=oem_v15.Agency(logo=inp_context.get("publisherLogo"))
                if "publisherLogo" in inp_context
                else None,
            )

        # filling the spatial section
        old_spatial = json_old.get("spatial")
        if old_spatial is None:
            spatial = None
        else:
            spatial = oem_v15.Spatial(
                location=old_spatial.get("location"),
                extent=old_spatial.get("extent"),
                resolution=old_spatial.get("resolution"),
            )

        # filling the temporal section
        inp_temporal = json_old.get("temporal")
        if inp_temporal is None:
            temporal = None
        else:
            inp_timeseries = inp_temporal.get("timeseries", [])
            if inp_timeseries is None:
                timeseries = None
            else:
                timeseries = [
                    oem_v15.Timeseries(
                        start=parse_date_or_none(ts.get("start")),
                        end=parse_date_or_none(ts.get("end")),
                        resolution=ts.get("resolution"),
                        ts_orientation=oem_v15.TimestampOrientation.create(
                            ts.get("alignment")
                        )
                        if "alignment" in ts and ts["alignment"] is not None
                        else None,
                        aggregation=ts.get("aggregationType"),
                    )
                    for ts in inp_timeseries
                ]
            temporal = oem_v15.Temporal(
                reference_date=parse_date_or_none(inp_temporal.get("referenceDate")),
                # TODO: does ** kwargs work on list?
                timeseries_collection=timeseries,
            )

        # filling the source section
        old_sources = json_old.get("sources")
        if old_sources is None:
            sources = None
        else:
            sources = [
                oem_v15.Source(
                    title=old_source.get("title"),
                    description=old_source.get("description"),
                    path=old_source.get("path"),
                    licenses=[
                        self.parse_term_of_use(l)
                        for l in old_source.get("licenses", [])
                    ],
                )
                for old_source in old_sources
            ]

        # filling the license section
        old_licenses = json_old.get("licenses")
        if old_licenses is None:
            licenses = None
        else:
            licenses = [
                self.parse_term_of_use(old_license) for old_license in old_licenses
            ]

        # filling the contributers section
        old_contributors = json_old.get("contributors")
        if old_contributors is None:
            contributors = None
        else:
            contributors = [
                oem_v15.Contribution(
                    contributor=oem_v15.Person(
                        name=old_contributor.get("title"),
                        email=old_contributor.get("email"),
                    ),
                    date=parse_date_or_none(old_contributor.get("date")),
                    obj=old_contributor.get("object"),
                    comment=old_contributor.get("comment"),
                )
                for old_contributor in old_contributors
            ]

        # extending with script-user information
        old_resources = json_old.get("resources")
        if old_resources is None:
            resources = None
        # Code added to raise exception when resource is empty
        else:
            if len(old_resources) == 0:
                raise ParserException("Resource field doesn't have any child entity")
            resources = []
            for resource in old_resources:
                old_schema = resource.get("schema")
                if old_schema is None:
                    schema = None
                else:
                    # filling the fields section
                    old_fields = old_schema.get("fields")
                    if old_fields is None:
                        fields = None
                    else:
                        fields = []

                        for field in old_fields:
                            # filling the is about section
                            old_is_abouts = field.get("isAbout")
                            if old_is_abouts is None:
                                is_about = None
                            else:
                                is_about = [
                                    oem_v15.IsAbout(
                                        name=old_is_about.get("name"),
                                        path=old_is_about.get("path"),
                                    )
                                    for old_is_about in old_is_abouts
                                ]

                            # filling the value reference section
                            old_value_references = field.get("valueReference")
                            if old_value_references is None:
                                value_reference = None
                            else:
                                value_reference = [
                                    oem_v15.ValueReference(
                                        value=old_value_reference.get("value"),
                                        name=old_value_reference.get("name"),
                                        path=old_value_reference.get("path"),
                                    )
                                    for old_value_reference in old_value_references
                                ]

                            fields.append(
                                oem_v15.Field(
                                    name=field.get("name"),
                                    description=field.get("description"),
                                    field_type=field.get("type"),
                                    is_about=is_about,
                                    value_reference=value_reference,
                                    unit=field.get("unit"),
                                )
                            )

                    field_dict = {field.name: field for field in fields or []}
                    old_foreign_keys = old_schema.get("foreignKeys", [])
                    foreign_keys = []
                    for fk in old_foreign_keys:
                        old_reference = fk.get("reference")
                        if old_reference is None:
                            raise ParserException("Foreign key without reference:", fk)
                        source_fields = [
                            field_dict[field_name]
                            for field_name in fk.get("fields", [])
                        ]
                        old_referenced_fields = old_reference.get("fields")
                        if old_referenced_fields is None:
                            referenced_fields = None
                        else:
                            referenced_fields = [
                                oem_v15.Field(name=fk_field)
                                for fk_field in old_referenced_fields
                            ]
                        referenced_resource = oem_v15.Resource(
                            name=old_reference.get("resource"),
                            schema=oem_v15.Schema(fields=referenced_fields),
                        )
                        for rf in referenced_fields:
                            rf.resource = referenced_resource
                        references = [
                            oem_v15.Reference(s, t)
                            for s, t in zip(source_fields, referenced_fields)
                        ]
                        foreign_keys.append(oem_v15.ForeignKey(references=references))
                    schema = oem_v15.Schema(
                        fields=fields,
                        primary_key=resource["schema"].get("primaryKey"),
                        foreign_keys=foreign_keys,
                    )
                old_dialect = resource.get("dialect")
                if old_dialect is None:
                    dialect = None
                else:
                    dialect = oem_v15.Dialect(
                        delimiter=resource["dialect"].get("delimiter"),
                        decimal_separator=resource["dialect"].get("decimalSeparator"),
                    )
                resources.append(
                    oem_v15.Resource(
                        profile=resource.get("profile"),
                        name=resource.get("name"),
                        path=resource.get("path"),
                        resource_format=resource.get("format"),
                        encoding=resource.get("encoding"),
                        schema=schema,
                        dialect=dialect,
                    )
                )

        inp_review = json_old.get("review")
        if inp_review is None:
            review = None
        else:
            review = oem_v15.Review(
                path=inp_review.get("path"), badge=inp_review.get("badge")
            )

        inp_comment = json_old.get("_comment")
        if inp_comment is None:
            comment = None
        else:
            comment = oem_v15.MetaComment(
                metadata_info=inp_comment.get("metadata"),
                dates=inp_comment.get("dates"),
                units=inp_comment.get("units"),
                languages=inp_comment.get("languages"),
                licenses=inp_comment.get("licenses"),
                review=inp_comment.get("review"),
                null=inp_comment.get("null"),
                todo=inp_comment.get("todo"),
            )

        metadata = oem_v15.OEPMetadata(
            name=json_old.get("name"),
            title=json_old.get("title"),
            identifier=json_old["id"],
            description=json_old.get("description"),
            subject=subject,
            languages=json_old.get("language"),
            keywords=json_old.get("keywords"),
            publication_date=parse_date_or_none(json_old.get("publicationDate")),
            context=context,
            spatial=spatial,
            temporal=temporal,
            sources=sources,
            terms_of_use=licenses,
            contributions=contributors,
            resources=resources,
            databus_identifier=json_old.get("@id"),
            databus_context=json_old.get("@context"),
            review=review,
            comment=comment,
        )
        return metadata

    def assert_1_5_metastring(self, json_string: str):
        """Checks string conformity to OEP Metadata Standard Version 1.5

        Parameters
        ----------
        json_string: str
            The JSON string to be checked.

        Returns
        -------
        bool
            True if valid, Raises Exception otherwise.
        """

        keys = [
            "title",
            "description",
            "language",
            "spatial",
            "temporal",
            "sources",
            "license",
            "contributions",
            "resources",
            "metadata_version",
        ]
        subkeys_spatial = ["location", "extent", "resolution"]
        subkeys_timeseries = [
            "start",
            "end",
            "resolution",
            "alignment",
            "aggregationType",
        ]
        subkeys_temporal = ["reference_date", "timeseries"]
        subkeys_license = ["id", "name", "version", "url", "instruction", "copyright"]
        object_subkeys = {
            "spatial": subkeys_spatial,
            "temporal": subkeys_temporal,
            "license": subkeys_license,
        }
        subkeys_sources = [
            "name",
            "description",
            "url",
            "license",
            "copyright",
        ]  # in list of objects
        subkeys_contributors = [
            "name",
            "email",
            "date",
            "comment",
        ]  # in list of objects
        subkeys_resources = ["name", "format", "fields"]  # in list of objects
        list_subkeys = {
            "sources": subkeys_sources,
            "contributions": subkeys_contributors,
            "resources": subkeys_resources,
        }
        subkeys_resources_fields = ["name", "description", "unit"]  # in list of objects

        json_dict = json.loads(json_string)
        try:
            # check if all top level keys are present
            for i in keys:
                if not i in json_dict.keys():
                    raise Exception(
                        'The String did not contain the key "{0}"'.format(i)
                    )
            # check for all keys in second level objects
            for key in object_subkeys:
                for subkey in object_subkeys[key]:
                    if not subkey in json_dict[key]:
                        raise Exception(
                            'The "{0}" object did not contain a "{1}" key'.format(
                                key, subkey
                            )
                        )
            # check for all objects in lists if they contain all required keys
            for key in list_subkeys:
                for list_element in json_dict[key]:
                    for subkey in list_subkeys[key]:
                        if not subkey in list_element:
                            raise Exception(
                                'An object in "{0}" is missing a "{1}" key'.format(
                                    key, subkey
                                )
                            )
        except Exception as error:
            print(
                "The input String does not conform to metadatastring version 1.3 standard"
            )
            print(error)

    # TODO make function check all subkeys as well
    def has_rogue_keys(self, json_string):
        """Checks all keys if they are part of the metadata specification. Gives warnings if not.

        Parameters
        ----------
        json_string: str
            The JSON string to be checked.

        Returns
        -------
        """

        json_dict = json.loads(json_string)
        allowed_keys = [
            "title",
            "description",
            "language",
            "spatial",
            "temporal",
            "sources",
            "license",
            "contributions",
            "resources",
            "metadata_version",
        ]
        for j in json_dict.keys():
            if not j in allowed_keys:
                print('Warning: "{0}" is not among the allowed keys'.format(j))

    def get_table_name(self, metadata_file):
        """Provides the tablename information from the metadata_file

        Parameters
        ----------
        metadata_file: str
            The metadata script from where the tablename is extracted from.

        Returns
        -------
        tablename: str
            returns the tablename.
        """

        raise NotImplementedError



from rdflib import Graph
from rdflib.compare import _squashed_graphs_triples
from rdflib.compare import isomorphic

from omi.dialects.oep.compiler import JSONCompiler
from omi.dialects.oep.parser import JSONParser_1_4
from omi.dialects.oep.renderer import JSONRenderer
from omi.dialects.rdf.compiler import RDFCompiler
from omi.dialects.rdf.parser import RDFParser


def test_roundtrip():
    json_compiler = JSONCompiler()
    json_parser = JSONParser_1_4()
    json_renderer = JSONRenderer()
    rdf_compiler = RDFCompiler()
    rdf_p = RDFParser()
    with open("tests/data/metadata_v14.ttl", "r") as _input_file:
        input_string = _input_file.read()
        expected_graph = Graph()
        expected_graph.parse(data=input_string, format="ttl")
        # Step 1: Parse Turtle to internal structure
        internal_metadata = rdf_p.parse_from_string(input_string)
        # Step 2: Translate to json string
        json_metadata = json_renderer.render(json_compiler.visit(internal_metadata))
        # Step 3: Parse json string
        internal_metadata2 = json_parser.parse_from_string(json_metadata)
        # Step 4: Translate to Turtle
        _ = rdf_compiler.visit(internal_metadata2)
        # Final step: Compare
        for (t1, t2) in _squashed_graphs_triples(expected_graph, rdf_compiler.graph):
            assert t1 == t2
        assert isomorphic(expected_graph, rdf_compiler.graph)



from omi.dialects.base.dialect import Dialect
from omi.dialects.base.register import register
from omi.dialects.oep.compiler import JSONCompiler
from omi.dialects.oep.compiler import JSONCompilerOEM15
from omi.dialects.oep.parser import JSONParser_1_3
from omi.dialects.oep.parser import JSONParser_1_4
from omi.dialects.oep.parser import JSONParser_1_5
from omi.dialects.oep.renderer import JSONRenderer
from omi.oem_structures import oem_v15


@register("oep-v1.3")
class OEP_V_1_3_Dialect(Dialect):
    _parser = JSONParser_1_3
    _compiler = None
    _renderer = JSONRenderer


@register("oep-v1.4")
class OEP_V_1_4_Dialect(Dialect):
    _parser = JSONParser_1_4
    _compiler = JSONCompiler
    _renderer = JSONRenderer


@register("oep-v1.5")
class OEP_V_1_5_Dialect(Dialect):
    _parser = JSONParser_1_5
    _compiler = JSONCompilerOEM15
    _renderer = JSONRenderer

    def compile(self, obj: oem_v15.OEPMetadata, *args, **kwargs):
        """
        Compiles the passed :class:`~omi.structure.OEPMetadata`-object into the
        structure fitting for this dialect

        Parameters
        ----------
        obj
            The :class:`~omi.structure.OEPMetadata`-object to compile

        Returns
        -------

        """
        c = self._compiler()
        return c.visit(obj, *args, **kwargs)

    def parse(self, string: str, *args, **kwargs) -> oem_v15.OEPMetadata:
        """
        Loads the passed string into an
        :class:`~omi.structure.OEPMetadata`-object.

        Parameters
        ----------
        string
            The string to parse

        Returns
        -------
            The :class:`~omi.structure.OEPMetadata`-object represented by `string`
        """
        p = self._parser()
        return p.parse_from_string(string, *args, **kwargs)

    def parse_file(self, path: str, *args, **kwargs) -> oem_v15.OEPMetadata:
        """
        Loads the passed string into an
        :class:`~omi.structure.OEPMetadata`-object.

        Parameters
        ----------
        string
            The string to parse

        Returns
        -------
            The :class:`~omi.structure.OEPMetadata`-object represented by `string`
        """
        p = self._parser()
        return p.parse_from_file(path, *args, **kwargs)

    def compile_and_render(self, obj: oem_v15.OEPMetadata, *args, **kwargs):
        """
        Combination of :func:`~omi.dialects.base.dialect.Dialect.compile` and :func:`~omi.dialects.base.dialect.Dialect.render`.
        """
        c = self._compiler()
        r = self._renderer()
        compiled = c.visit(obj, *args, **kwargs)
        return r.render(compiled, *args, **kwargs)



from omi import structure
from omi.oem_structures import oem_v15


class Compiler:
    """
    Compiles :class:`~omi.structure.Compilable` objects into the respective metadata
    format. Every omi compiler should inherit from this class
    """

    def visit(self, obj, *args, **kwargs):
        """
        Calls the respective compiler for :class:`~omi.structure.Compilable` objects
        respective to :attr:`Compilable.__compiler_name__`


        Parameters
        ----------
        obj
            Object to compile

        Returns
        -------
            Metadata representation of `obj`

        """
        if isinstance(obj, list):
            return [self.visit(x) for x in obj if x is not None]
        if isinstance(obj, structure.Compilable):
            meth = getattr(self, "visit_{name}".format(name=obj.__compiler_name__))
            return meth(obj, *args, **kwargs)
        else:
            return obj

    def visit_context(self, context: structure.Context, *args, **kwargs):
        raise NotImplementedError

    def visit_contributor(self, contributor: structure.Contribution, *args, **kwargs):
        raise NotImplementedError

    def visit_language(self, language: structure.Language, *args, **kwargs):
        raise NotImplementedError

    def visit_spatial(self, spatial: structure.Spatial, *args, **kwargs):
        raise NotImplementedError

    def visit_temporal(self, temporal: structure.Temporal, *args, **kwargs):
        raise NotImplementedError

    def visit_timestamp_orientation(
        self, ts_orientation: structure.TimestampOrientation, *args, **kwargs
    ):
        raise NotImplementedError

    def visit_timeseries(self, timeseries: oem_v15.Timeseries, *args, **kwargs):
        raise NotImplementedError

    def visit_source(self, source: structure.Source, *args, **kwargs):
        raise NotImplementedError

    def visit_license(self, lic: structure.License, *args, **kwargs):
        raise NotImplementedError

    def visit_resource(self, resource: structure.Resource, *args, **kwargs):
        raise NotImplementedError

    def visit_schema(self, context: structure.Schema, *args, **kwargs):
        raise NotImplementedError

    def visit_dialect(self, context: structure.Dialect, *args, **kwargs):
        raise NotImplementedError

    def visit_field(self, field: structure.Field, *args, **kwargs):
        raise NotImplementedError

    def visit_foreign_key(self, foreign_key: structure.ForeignKey, *args, **kwargs):
        raise NotImplementedError

    def visit_reference(self, reference: structure.Reference, *args, **kwargs):
        raise NotImplementedError

    def visit_review(self, review: structure.Review, *args, **kwargs):
        raise NotImplementedError

    def visit_meta_comment(self, comment: structure.MetaComment, *args, **kwargs):
        raise NotImplementedError

    def visit_metadata(self, metadata: structure.OEPMetadata, *args, **kwargs):
        raise NotImplementedError



from rdflib import BNode
from rdflib import Graph
from rdflib.compare import _squashed_graphs_triples
from rdflib.compare import graph_diff
from rdflib.compare import isomorphic
from rdflib.compare import similar
from rdflib.compare import to_isomorphic

from omi.dialects.rdf.compiler import RDFCompiler

from ..internal_structures import metadata_v_1_4
from omi.dialects.rdf.namespace import OEO
from rdflib.namespace import DCTERMS


def test_compiler_v1_4():
    compiler = RDFCompiler()
    with open("tests/data/metadata_v14.ttl", "r", encoding="utf-8") as _input_file:
        expected_graph = Graph()
        expected_graph.parse(data=_input_file.read(), format="ttl")
        _ = compiler.visit(metadata_v_1_4)
        expected = to_isomorphic(expected_graph)
        got = to_isomorphic(compiler.graph)
        for (t1, t2) in _squashed_graphs_triples(expected, got):
            assert t1 == t2
        assert isomorphic(expected, got)



from typing import Dict
from typing import Iterable
from typing import Tuple

from dateutil.parser import parse as parse_date
from rdflib import Graph
from rdflib.graph import Node
from rdflib.graph import URIRef
from rdflib.namespace import DCTERMS
from rdflib.namespace import FOAF
from rdflib.namespace import RDF
from rdflib.namespace import RDFS

from omi import structure as struc
from omi.dialects.base.parser import Parser
from omi.dialects.rdf.namespace import ADMS
from omi.dialects.rdf.namespace import DCAT
from omi.dialects.rdf.namespace import DCATDE
from omi.dialects.rdf.namespace import OEO
from omi.dialects.rdf.namespace import SCHEMA
from omi.dialects.rdf.namespace import SKOS
from omi.dialects.rdf.namespace import SPDX


def _only(gen):
    r = _one_or_none(gen)
    if r is None:
        raise Exception("No matching elements")
    return r


def _one_str_or_none(gen):
    r = _one_or_none(gen)
    if r is None:
        return None
    return str(r)


def _one_or_none(gen):
    l = list(gen)
    if not l:
        return None
    if len(l) > 1:
        raise Exception("Found more than one match:" + str(l))
    return l[0]


class RDFParser(Parser):
    def load_string(self, string: str, *args, **kwargs):
        g = Graph()
        g.parse(data=string, format="ttl")
        return g

    def parse(self, graph, *args, **kwargs):
        for dataset in {s for s, _, _ in graph}.difference({o for _, _, o in graph}):
            if dataset in graph.subjects(RDF.type, DCAT.Dataset):
                return self.parse_metadata(graph, dataset)

    def parse_date(self, node: Node):
        return parse_date(node)

    def parse_context(self, graph: Graph, parent: Node) -> struc.Context:
        fa = _only(graph.objects(parent, OEO.has_funding_agency))
        kwargs = {}
        if fa is not None:
            kwargs["funding_agency"] = struc.Agency(
                name=_one_str_or_none(graph.objects(fa, DCTERMS.title)),
                logo=_one_str_or_none(graph.objects(fa, OEO.has_logo)),
            )
        pa = _only(graph.objects(parent, OEO.has_publisher))
        if pa is not None:
            kwargs["publisher"] = struc.Agency(
                name=_one_str_or_none(graph.objects(pa, DCTERMS.title)),
                logo=_one_str_or_none(graph.objects(pa, OEO.has_logo)),
            )
        return struc.Context(
            contact=_one_str_or_none(graph.objects(parent, DCAT.contactpoint)),
            documentation=_one_str_or_none(graph.objects(parent, OEO.documentation)),
            grant_number=_one_str_or_none(graph.objects(parent, OEO.grantNo)),
            homepage=_one_str_or_none(graph.objects(parent, FOAF.homepage)),
            source_code=_one_str_or_none(graph.objects(parent, OEO.sourceCode)),
            **kwargs
        )

    def parse_contributor(self, graph: Graph, parent: Node) -> struc.Contribution:
        return struc.Contribution(
            contributor=self.parse_person(
                graph, _only(graph.objects(parent, DCTERMS.contributor))
            ),
            date=self.parse_date(_only(graph.objects(parent, OEO.date))),
            obj=_one_str_or_none(graph.objects(parent, OEO.object)),
            comment=_one_str_or_none(graph.objects(parent, OEO.comment)),
        )

    def parse_person(self, graph, parent: Node):
        return struc.Person(
            name=_one_str_or_none(graph.objects(parent, FOAF.name)),
            email=_one_str_or_none(graph.objects(parent, FOAF.mbox)),
        )

    def parse_spatial(self, graph: Graph, parent: Node) -> struc.Spatial:
        return struc.Spatial(
            extent=_one_str_or_none(graph.objects(parent, SKOS.prefLabel)),
            location=_one_str_or_none(graph.objects(parent, OEO.location)),
            resolution=_one_str_or_none(
                graph.objects(parent, OEO.has_spatial_resolution)
            ),
        )

    def parse_temporal(self, graph: Graph, parent: Node) -> struc.Temporal:

        orientation = self.parse_timestamp_orientation(
            _only(graph.objects(parent, OEO.has_timestamp_alignment))
        )

        return struc.Temporal(
            start=self.parse_date(_only(graph.objects(parent, SCHEMA.startDate))),
            end=self.parse_date(_only(graph.objects(parent, SCHEMA.endDate))),
            ts_orientation=orientation,
            reference_date=self.parse_date(
                _only(graph.objects(parent, OEO.referenceDate))
            ),
            resolution=_one_str_or_none(graph.objects(parent, OEO.has_time_resolution)),
            aggregation=_one_str_or_none(graph.objects(parent, OEO.uses_aggregation)),
        )

    def parse_timestamp_orientation(self, node):
        if node == OEO.left_orientation:
            return struc.TimestampOrientation.left
        elif node == OEO.middle_orientation:
            return struc.TimestampOrientation.middle
        elif node == OEO.right_orientation:
            return struc.TimestampOrientation.right
        else:
            raise Exception("Unknown timestamp orientation: {}".format(node))

    def parse_source(self, graph: Graph, parent: Node) -> struc.Source:
        return struc.Source(
            title=_one_str_or_none(graph.objects(parent, DCTERMS.title)),
            description=_one_str_or_none(graph.objects(parent, DCTERMS.description)),
            path=_one_str_or_none(graph.objects(parent, FOAF.page)),
            licenses=[
                self.parse_terms_of_use(graph, tos)
                for tos in graph.objects(parent, OEO.has_terms_of_use)
            ],
        )

    def parse_terms_of_use(self, graph: Graph, parent: Node) -> struc.TermsOfUse:
        if isinstance(parent, URIRef):
            return None
        else:

            return struc.TermsOfUse(
                lic=self.parse_license(
                    graph, _only(graph.objects(parent, DCAT.license))
                ),
                attribution=str(
                    _only(graph.objects(parent, DCATDE.licenseAttributionByText))
                ),
                instruction=_one_str_or_none(
                    graph.objects(parent, OEO.has_instruction)
                ),
            )

    def parse_license(self, graph, parent: Node):
        kw = dict()
        for c in graph.objects(parent, RDFS.comment):
            kw["comment"] = _one_str_or_none(c)
        refs = list(graph.objects(parent, RDFS.seeAlso))
        return struc.License(
            name=_one_str_or_none(graph.objects(parent, SPDX.name)),
            identifier=_one_str_or_none(graph.objects(parent, SPDX.licenseId)),
            path=_one_str_or_none(graph.objects(parent, FOAF.page)),
            other_references=refs if refs else None,
            text=_one_str_or_none(graph.objects(parent, SPDX.licenseText)),
            **kw
        )

    def parse_resource(
        self,
        graph: Graph,
        parent: Node,
        resources: Dict[str, Tuple[struc.Resource, Dict[str, struc.Field]]] = None,
    ) -> struc.Resource:
        rname = _one_str_or_none(graph.objects(parent, DCTERMS.title))
        resources = resources or dict()
        if resources and rname in resources:
            return resources[rname][0]
        else:
            dialect_node = _one_or_none(graph.objects(parent, OEO.has_dialect))
            dialect = self.parse_dialect(graph, dialect_node) if dialect_node else None
            r = struc.Resource(
                dialect=dialect,
                encoding=_one_str_or_none(graph.objects(parent, OEO.encoding)),
                name=rname,
                path=_one_str_or_none(graph.objects(parent, DCAT.accessURL)),
                profile=_one_str_or_none(graph.objects(parent, OEO.profile)),
                resource_format=_one_str_or_none(graph.objects(parent, OEO.has_format)),
                schema=None,
            )
            r_fields = {}
            resources[rname] = r, r_fields
            schema = self.parse_schema(graph, parent, resources=resources)
            r.schema = schema
            for f in schema.fields:
                if f.name not in r_fields:
                    r_fields[f.name] = f
            return r

    def parse_schema(
        self,
        graph: Graph,
        parent: Node,
        resources: Dict[str, Tuple[Node, Iterable[Node]]] = None,
    ) -> struc.Schema:
        return struc.Schema(
            fields=[
                self.parse_field(graph, f) for f in graph.objects(parent, OEO.field)
            ],
            primary_key=[
                self.parse_field(graph, f).name
                for f in graph.objects(parent, OEO.primaryKey)
            ],
            foreign_keys=[
                self.parse_foreign_key(graph, f, resources=resources)
                for f in graph.objects(parent, OEO.has_foreignKey)
            ],
        )

    def parse_dialect(self, graph: Graph, parent: Node) -> struc.Dialect:
        delim = _one_or_none(graph.objects(parent, DCTERMS.delimiter))
        if delim is not None:
            delim = str(delim)
        return struc.Dialect(
            decimal_separator=_one_str_or_none(
                graph.objects(parent, OEO.decimalSeparator)
            ),
            delimiter=delim,
        )

    def parse_field(self, graph: Graph, parent: Node) -> struc.Field:
        return struc.Field(
            name=_one_str_or_none(graph.objects(parent, DCTERMS.title)),
            unit=_one_str_or_none(graph.objects(parent, OEO.unit)),
            field_type=_one_str_or_none(graph.objects(parent, OEO.type)),
            description=_one_str_or_none(graph.objects(parent, DCTERMS.description)),
        )

    def parse_foreign_key(
        self,
        graph: Graph,
        parent: Node,
        resources: Dict[str, Tuple[Node, Iterable[Node]]] = None,
    ) -> struc.ForeignKey:
        return struc.ForeignKey(
            references=[
                self.parse_reference(graph, r)
                for r in graph.objects(parent, OEO.has_reference)
            ]
        )

    def parse_reference(
        self,
        graph: Graph,
        parent: Node,
        resources: Dict[str, Tuple[Node, Iterable[Node]]] = None,
    ) -> struc.Reference:
        target_node = _only(graph.objects(parent, OEO.has_target))
        target_field = self.parse_field(graph, target_node)
        target_resource = self.parse_resource(
            graph, _only(graph.subjects(OEO.field, target_node))
        )
        target_field.resource = target_resource
        return struc.Reference(
            source=self.parse_field(
                graph, _only(graph.objects(parent, OEO.has_source))
            ),
            target=target_field,
        )

    def parse_review(self, graph: Graph, parent: Node) -> struc.Review:
        return struc.Review(
            badge=_one_str_or_none(graph.objects(parent, OEO.has_badge)),
            path=_one_str_or_none(graph.objects(parent, FOAF.page)),
        )

    def parse_meta_comment(self, graph: Graph, parent: Node) -> struc.MetaComment:
        return struc.MetaComment(
            dates=_one_str_or_none(graph.objects(parent, OEO.dates_info)),
            languages=_one_str_or_none(graph.objects(parent, OEO.languages_info)),
            licenses=_one_str_or_none(graph.objects(parent, OEO.licenses_info)),
            metadata_info=_one_str_or_none(graph.objects(parent, OEO.metadata_info)),
            none=_one_str_or_none(graph.objects(parent, OEO.none_info)),
            review=_one_str_or_none(graph.objects(parent, OEO.review_info)),
            units=_one_str_or_none(graph.objects(parent, OEO.units_info)),
        )

    def parse_metadata(self, graph: Graph, parent: Node) -> struc.OEPMetadata:
        context = self.parse_context(graph, parent)
        contributors = [
            self.parse_contributor(graph, c)
            for c in graph.objects(parent, OEO.has_contribution)
        ]
        language = [str(l) for l in graph.objects(parent, DCTERMS.language)]
        spatial = self.parse_spatial(
            graph, _only(graph.objects(parent, DCTERMS.spatial))
        )

        temporal = self.parse_temporal(
            graph, _only(graph.objects(parent, DCTERMS.temporal))
        )
        comment = self.parse_meta_comment(
            graph, _only(graph.objects(parent, OEO.comment))
        )
        resources = [
            self.parse_resource(graph, r)
            for r in graph.objects(parent, OEO.has_resource)
        ]
        terms_of_use = [
            self.parse_terms_of_use(graph, l)
            for l in graph.objects(parent, OEO.has_terms_of_use)
        ]
        sources = [
            self.parse_source(graph, s) for s in graph.objects(parent, DCTERMS.source)
        ]
        review = self.parse_review(graph, _only(graph.objects(parent, OEO.has_review)))
        return struc.OEPMetadata(
            comment=comment,
            context=context,
            contributions=contributors,
            description=_one_str_or_none(graph.objects(parent, DCTERMS.description)),
            identifier=str(parent),
            keywords=list(map(str, graph.objects(parent, DCAT.keyword))),
            languages=language,
            name=_one_str_or_none(graph.objects(parent, ADMS.Identifier)),
            terms_of_use=terms_of_use,
            publication_date=self.parse_date(
                _only(graph.objects(parent, OEO.publicationDate))
            ),
            resources=resources,
            review=review,
            sources=sources,
            spatial=spatial,
            temporal=temporal,
            title=_one_str_or_none(graph.objects(parent, DCTERMS.title)),
        )



import json
from collections import OrderedDict

from omi.dialects.base.renderer import Renderer


class JSONRenderer(json.JSONEncoder, Renderer):
    """
    This enconder sets up a structured oder of the json string when transforming
    it from a python OrderedDict
    """

    def __init__(self, *args, **kwargs):
        super(JSONRenderer, self).__init__(*args, **kwargs)
        self.current_indent = 0
        self.current_indent_str = ""

    def encode(self, o):
        # Special Processing for lists
        if isinstance(o, (list, tuple)):
            primitives_only = True
            for item in o:
                if isinstance(item, (list, tuple, OrderedDict)):
                    primitives_only = False
                    break
            output = []
            if primitives_only:
                for item in o:
                    output.append(json.dumps(item, ensure_ascii=False))
                return "[ " + ", ".join(output) + "  ]"
            else:
                self.current_indent += 2
                self.current_indent_str = "".join(
                    [" " for x in range(self.current_indent)]
                )
                liste = []
                for item in o:
                    output = []
                    # This is performed if in the list is a OrderedDict
                    if isinstance(item, OrderedDict):
                        for key, value in item.items():
                            output.append(json.dumps(key) + ": " + self.encode(value))

                        liste.append(
                            "\n"
                            + 2 * self.current_indent_str
                            + "{"
                            + (",\n" + 2 * self.current_indent_str).join(output)
                            + "}"
                        )

                    else:
                        raise AssertionError(
                            "Only OrderedDicts in lists are properly structured. Please redefine it in the encode function."
                        )
                        output.append(self.current_indent_str + self.encode(item))
                        return "[\n" + ",".join(output) + "]"

                self.current_indent -= 2
                self.current_indent_str = "".join(
                    [" " for x in range(self.current_indent)]
                )

            return "[" + ",".join(liste) + "]"

        elif isinstance(o, OrderedDict):
            output = []
            self.current_indent += 4
            self.current_indent_str = "".join([" " for x in range(self.current_indent)])
            for key, value in o.items():
                output.append(
                    self.current_indent_str
                    + json.dumps(key)
                    + ": "
                    + self.encode(value)
                )
            self.current_indent -= 4
            self.current_indent_str = "".join([" " for x in range(self.current_indent)])
            return "{\n" + ",\n".join(output) + "}"
        else:
            return json.dumps(o, ensure_ascii=False)

    def render(self, inp, *args, **kwargs):
        return self.encode(inp)



from omi.dialects.base.compiler import Compiler
from omi.dialects.base.parser import Parser
from omi.dialects.base.renderer import Renderer
from omi.structure import OEPMetadata


class Dialect:
    _parser = Parser
    _compiler = Compiler
    _renderer = Renderer

    def compile(self, obj: OEPMetadata, *args, **kwargs):
        """
        Compiles the passed :class:`~omi.structure.OEPMetadata`-object into the
        structure fitting for this dialect

        Parameters
        ----------
        obj
            The :class:`~omi.structure.OEPMetadata`-object to compile

        Returns
        -------

        """
        c = self._compiler()
        return c.visit(obj, *args, **kwargs)

    def parse(self, string: str, *args, **kwargs) -> OEPMetadata:
        """
        Loads the passed string into an
        :class:`~omi.structure.OEPMetadata`-object.

        Parameters
        ----------
        string
            The string to parse

        Returns
        -------
            The :class:`~omi.structure.OEPMetadata`-object represented by `string`
        """
        p = self._parser()
        return p.parse_from_string(string, *args, **kwargs)

    def render(self, structure, *args, **kwargs) -> str:
        """
        Turns the structure used by this dialect into a string

        Parameters
        ----------
        structure
            The structure to stringify

        Returns
        -------
            A string representation of `structure`
        """
        r = self._renderer()
        return r.render(structure, *args, **kwargs)

    def compile_and_render(self, obj: OEPMetadata, *args, **kwargs):
        """
        Combination of :func:`~omi.dialects.base.dialect.Dialect.compile` and :func:`~omi.dialects.base.dialect.Dialect.render`.
        """
        c = self._compiler()
        r = self._renderer()
        compiled = c.visit(obj, *args, **kwargs)
        return r.render(compiled, *args, **kwargs)



"""
This module extends the internal data structure 
with the latest changes of the OEMetadata standard.

An OEMeadata version specific dialect can import it.

This bypasses the limitation of OMI to rely on a single 
internal representation of the data structure, since 
OEMetadata is not a static structure and OMI must be
be able to validate multiple OEMetadata versions.
"""
# TODO: maybe change all inheritance to omi.structure to
#      avoide alot of redundancy

from datetime import datetime
from enum import Enum
from typing import Iterable
from omi.structure import Compilable


class Language(Compilable):
    __compiler_name__ = "language"


class Subject(Compilable):
    __compiler_name__ = "subject"

    def __init__(self, name: str = None, path: str = None):
        self.name = name
        self.path = path


class Spatial(Compilable):
    __compiler_name__ = "spatial"

    def __init__(
        self, location: str = None, extent: str = None, resolution: str = None
    ):
        self.location = location
        self.extent = extent
        self.resolution = resolution


class TimestampOrientation(Compilable, Enum):
    __compiler_name__ = "timestamp_orientation"
    left = 0
    middle = 1
    right = 2

    @staticmethod
    def create(value: str) -> "TimestampOrientation":
        if value == "left":
            return TimestampOrientation.left
        elif value == "middle":
            return TimestampOrientation.middle
        elif value == "right":
            return TimestampOrientation.right
        else:
            raise Exception("Unknown timestamp orientation:", value)


class Timeseries(Compilable):
    __compiler_name__ = "timeseries"

    def __init__(
        self,
        start: datetime = None,
        end: datetime = None,
        resolution: str = None,
        ts_orientation: TimestampOrientation = None,
        aggregation: str = None,
    ):
        self.ts_start = start
        self.ts_end = end
        self.ts_resolution = resolution
        self.ts_orientation = ts_orientation
        self.aggregation = aggregation


class Temporal(Compilable):
    __compiler_name__ = "temporal"

    def __init__(
        self,
        reference_date: datetime = None,
        timeseries_collection: Iterable[Timeseries] = None,
    ):
        self.reference_date = reference_date
        self.timeseries_collection = timeseries_collection


class License(Compilable):
    __compiler_name__ = "license"

    def __init__(
        self,
        name: str = None,
        identifier: str = None,
        text: str = None,
        path: str = None,
        other_references: Iterable[str] = None,
        comment: str = None,
    ):
        self.name = name
        self.path = path
        self.identifier = identifier
        self.other_references = other_references
        self.text = text
        self.comment = comment

    @staticmethod
    def instance_name_from_id(identifier: str):
        return "L_" + identifier.replace("-", "_").replace(".", "_").replace(
            "+", "_plus"
        )


class TermsOfUse(Compilable):
    __compiler_name__ = "terms_of_use"

    def __init__(
        self, instruction: str = None, attribution: str = None, lic: License = None
    ):
        self.instruction = instruction
        self.attribution = attribution
        self.license = lic


class Source(Compilable):
    __compiler_name__ = "source"

    def __init__(
        self,
        title: str = None,
        description: str = None,
        path: str = None,
        licenses: Iterable[TermsOfUse] = None,
    ):
        self.title = title
        self.description = description
        self.path = path
        self.licenses = licenses


class Person(Compilable):
    __compiler_name__ = "person"

    def __init__(self, name: str = None, email: str = None):
        self.name = name
        self.email = email


class Contribution(Compilable):
    __compiler_name__ = "contribution"

    def __init__(
        self,
        contributor: Person = None,
        date: datetime = None,
        obj: str = None,
        comment: str = None,
    ):
        self.contributor = contributor
        self.date = date
        self.object = obj
        self.comment = comment


class IsAbout(Compilable):
    __compiler_name__ = "is_about"

    def __init__(self, name: str = None, path: str = None):
        self.name = name
        self.path = path


class ValueReference(Compilable):
    __compiler_name__ = "value_reference"

    def __init__(self, value: str = None, name: str = None, path: str = None):
        self.value = value
        self.name = name
        self.path = path


class Field(Compilable):
    __compiler_name__ = "field"

    def __init__(
        self,
        name: str = None,
        description: str = None,
        field_type: str = None,
        is_about: Iterable[IsAbout] = None,
        value_reference: Iterable[ValueReference] = None,
        unit: str = None,
        resource: "Resource" = None,
    ):
        self.name = name
        self.description = description
        self.type = field_type
        self.is_about = is_about
        self.value_reference = value_reference
        self.unit = unit
        self.resource = resource

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ",".join(
                "{}={}".format(key, val)
                for key, val in self.__dict__.items()
                if key != "resource"
            ),
        )


class Agency(Compilable):
    __compiler_name__ = "agency"

    def __init__(self, name: str = None, logo: str = None):
        self.name = name
        self.logo = logo


class Context(Compilable):
    __compiler_name__ = "context"

    def __init__(
        self,
        homepage: str = None,
        documentation: str = None,
        source_code: str = None,
        contact: str = None,
        grant_number: str = None,
        funding_agency: Agency = None,
        publisher: Agency = None,
    ):
        self.homepage = homepage
        self.documentation = documentation
        self.source_code = source_code
        self.contact = contact
        self.grant_number = grant_number
        self.funding_agency = funding_agency
        self.publisher = publisher


class Reference(Compilable):
    __compiler_name__ = "reference"

    def __init__(self, source: Field = None, target: Field = None):
        self.source = source
        self.target = target


class ForeignKey(Compilable):
    __compiler_name__ = "foreign_key"

    def __init__(self, references: Iterable[Reference] = None):
        self.references = references


class Schema(Compilable):
    __compiler_name__ = "schema"

    def __init__(
        self,
        fields: Iterable[Field] = None,
        primary_key: Iterable[str] = None,
        foreign_keys: Iterable[ForeignKey] = None,
    ):
        self.fields = fields
        self.primary_key = primary_key
        self.foreign_keys = foreign_keys


class Dialect(Compilable):
    __compiler_name__ = "dialect"

    def __init__(self, delimiter: str = None, decimal_separator: str = None):
        self.delimiter = delimiter
        self.decimal_separator = decimal_separator


class Resource(Compilable):
    __compiler_name__ = "resource"

    def __init__(
        self,
        name: str = None,
        path: str = None,
        profile: str = None,
        resource_format: str = None,
        encoding: str = None,
        schema: Schema = None,
        dialect: Dialect = None,
    ):
        self.name = name
        self.path = path
        self.profile = profile
        self.format = resource_format
        self.encoding = encoding
        self.schema = schema
        if schema is not None:
            for field in schema.fields:
                field.resource = self
        self.dialect = dialect


class MetaComment(Compilable):
    __compiler_name__ = "meta_comment"

    def __init__(
        self,
        metadata_info: str = None,
        dates: str = None,
        units: str = None,
        languages: str = None,
        licenses: str = None,
        review: str = None,
        null: str = None,
        todo: str = None,
    ):
        self.metadata_info = metadata_info
        self.dates = dates
        self.units = units
        self.languages = languages
        self.licenses = licenses
        self.review = review
        self.null = null
        self.todo = todo


class Review(Compilable):
    __compiler_name__ = "review"

    def __init__(self, path: str = None, badge: str = None):
        self.path = path
        self.badge = badge


class OEPMetadata(Compilable):
    __compiler_name__ = "metadata"
    __required__ = ["id"]

    def __init__(
        self,
        name: str = None,
        title: str = None,
        identifier: str = None,
        description: str = None,
        languages: Iterable[Language] = None,
        subject: Iterable[Subject] = None,
        keywords: Iterable[str] = None,
        publication_date: datetime = None,
        context: Context = None,
        spatial: Spatial = None,
        temporal: Temporal = None,
        sources: Iterable[Source] = None,
        terms_of_use: Iterable[TermsOfUse] = None,
        contributions: Iterable[Contribution] = None,
        resources: Iterable[Resource] = None,
        databus_identifier: str = None,
        databus_context: str = None,
        review: Review = None,
        comment: MetaComment = None,
    ):
        self.name = name
        self.title = title
        self.identifier = identifier
        self.description = description
        self.subject = subject
        self.languages = languages
        self.keywords = keywords
        self.publication_date = publication_date
        self.context = context
        self.spatial = spatial
        self.temporal = temporal
        self.sources = sources
        self.license = terms_of_use
        self.contributions = contributions
        self.resources = resources
        self.databus_identifier = databus_identifier
        self.databus_context = databus_context
        self.review = review
        self.comment = comment

    def has_keywords(self):
        return self.keywords is not None



from omi.dialects.base.dialect import Dialect
from omi.dialects.base.register import register
from omi.dialects.rdf.compiler import RDFCompiler
from omi.dialects.rdf.parser import RDFParser
from omi.dialects.rdf.renderer import GraphRenderer
from omi.structure import OEPMetadata


@register("oep-rdf-v1.4")
class OEP_V_1_4_RDF_Dialect(Dialect):
    _parser = RDFParser
    _compiler = RDFCompiler
    _renderer = GraphRenderer

    def compile_and_render(self, obj: OEPMetadata, *args, **kwargs):
        c = self._compiler()
        r = self._renderer()
        c.visit(obj, *args, **kwargs)
        return r.render(c.graph, *args, **kwargs).decode("utf-8")



from omi.dialects.base.dialect import Dialect

DIALECT_DICT = {}


def register(identifier: str):
    def inner(dialect_class: Dialect):
        DIALECT_DICT[identifier] = dialect_class
        return dialect_class

    return inner


def get_dialect(identifier: str) -> Dialect:
    return DIALECT_DICT[identifier]



from typing import TypeVar

from omi.structure import OEPMetadata

T = TypeVar("T")


class ParserException(Exception):
    pass


class Parser:
    """
    A parser is used to transform to read a specific metadata format and
    transform it into the internal metadata representation.
    """
    def parse(self, structure: T, *args, **kwargs) -> OEPMetadata:
        """
        Transforms the input structure into metadata as used by the
        OpenEnergyPlatform

        Parameters
        ----------
        inp: str
            The input string that should be parsed into OEP metadata

        Returns
        -------
        OEPMetadata
            OEP metadata represented by `inp`
        """
        raise NotImplementedError

    def load_string(self, string: str, *args, **kwargs):
        """
        Load a string into the structure represented by the dialect
        Parameters
        ----------
        string: str

        Returns
        -------
            Translates the passed string into the format used as input for this parser
        """
        raise NotImplementedError

    def parse_from_string(
        self,
        string: str,
        load_args=None,
        parse_args=None,
        load_kwargs=None,
        parse_kwargs=None,
    ) -> OEPMetadata:
        """
        Parse a string into :class:`~omi.structure.OEPMetadata`

        Parameters
        ----------
        string

        Returns
        -------

        """
        return self.parse(
            self.load_string(string, *(load_args or []), **(load_kwargs or {})),
            *(parse_args or []),
            **(parse_kwargs or {})
        )

    @staticmethod
    def __unpack_file(file, **kwargs):
        """

        Parameters
        ----------

        Returns
        -------

        """
        with open(file, **kwargs, encoding='utf-8') as inp:
            return inp.read()

    def parse_from_file(self, file_path, *args, **kwargs):
        return self.parse_from_string(self.__unpack_file(file_path), *args, **kwargs)

    def is_valid(self, inp: str) -> bool:
        """
        Verify whether `inp` is a sting representation that is parsable by this
        parser

        Parameters
        ----------
        inp: str
            String to verify

        Returns
        -------
        bool:
            Indicated whether this object is parsable or not

        """
        raise NotImplementedError

    def is_file_valid(self, file: str, **kwargs):
        """
        Verify whether the contents of the file under `file` is parsable by this
        parser

        Parameters
        ----------
        file: str
            Path to the file to validate

        **kwargs:

        Returns
        -------
        bool:
            Returns `True` iff the file's content is parsable

        """
        return self.is_valid(self.__unpack_file(file, **kwargs))



class Renderer:
    def render(self, inp, *args, **kwargs) -> str:
        """
        Transforms the given structure string into a string

        Parameters
        ----------
        inp
            The structure to render

        Returns
        -------
        str
            A string representation of `inp`
        """
        raise NotImplementedError

    @staticmethod
    def __unpack_file(*args, **kwargs):
        """

        Parameters
        ----------

        Returns
        -------

        """
        with open(*args, **kwargs) as inp:
            return inp.read()

    def render_to_file(self, inp, file_path, *args, **kwargs):
        with open(file_path, "w") as outfile:
            outfile.write(self.render(inp, *args, **kwargs))



import datetime
import imp
import unittest

from omi.structure import Compilable
from omi.structure import Field
from omi.oem_structures import oem_v15


def _test_generic_parsing(parser, inp, expected, **kwargs):
    if parser:
        result = parser.parse_from_file(inp)
        assert_compileable_equal(expected, result, **kwargs)


def assert_compileable_equal(expected, got, nulls=None, exclude=None):
    exclude = exclude or []
    nulls = nulls or [None]
    assert isinstance(expected, (Compilable, dict)), type(expected)
    assert isinstance(got, (Compilable, dict)), (type(got), expected)
    for key in set(expected.__dict__.keys()).union(set(got.__dict__.keys())):
        if (
            key not in exclude
            and isinstance(getattr(expected, key), Field)
            or isinstance(getattr(expected, key), oem_v15.Field)
        ):
            l = getattr(expected, key)
            r = getattr(got, key)
            if isinstance(l, Compilable):
                new_exclude = []
                if isinstance(l, Field):
                    new_exclude = ["resource"]
                elif isinstance(l, oem_v15.Field):
                    new_exclude = ["resource"]
                assert_compileable_equal(l, r, nulls=nulls, exclude=new_exclude)
            elif isinstance(l, list) and isinstance(r, list):
                assert len(l) == len(
                    r
                ), "Lists do not match (Expected: {}; Got: {})".format(l, r)
                for l0, r0 in zip(sorted(l), sorted(r)):
                    if isinstance(l0, Compilable):
                        if isinstance(l0, Field):
                            new_exclude = ["resource"]
                        elif isinstance(l, oem_v15.Field):
                            new_exclude = ["resource"]
                        else:
                            new_exclude = []
                        assert_compileable_equal(
                            l0, r0, nulls=nulls, exclude=new_exclude
                        )
                    else:
                        assert l0 == r0, "Expected: {}; Got: {}".format(l0, r0)
            else:
                if not (l is None and r in nulls):
                    assert (
                        l == r
                    ), "Keys {} do not match (Expected:{}:{}; Got: {}:{})".format(
                        key, type(l), l, type(r), r
                    )


def parse_date(s):
    return datetime.datetime.strptime(s, "%Y-%M-%d")



# -*- coding: utf-8 -*-

from omi.dialects.rdf.parser import RDFParser

from ..base.parser import _test_generic_parsing
from ..internal_structures import metadata_v_1_4


def test_parser_rdf():
    parser = RDFParser()
    _input_file = "tests/data/metadata_v14.ttl"
    expected_result = metadata_v_1_4
    _test_generic_parsing(parser, _input_file, expected_result)



import datetime

from omi import structure
from omi.dialects.oep.compiler import JSONCompiler
from omi.dialects.oep.parser import JSONParser_1_3


def metadata_conversion(old_sql, new_sql, user, user_email):
    """ Conversion of an existing metadata file to a newer version

    Parameters
    ----------

    old_sql: str
        The path to the file containing the old sql file.
    new_sql: str
        The filename of the new sql file.
    user: str
        The name of the user for the 'contributions' section
    user_email: str
        The email address of the user.

    Returns
    -------
    """

    parser = JSONParser_1_3()
    metadata = parser.parse_from_file(old_sql)

    metadata.contributors.append(
        structure.Contribution(
            title=user,
            email=user_email,
            date=datetime.now,
            obj=None,
            comment="Update metadata to v1.3 using metadata conversion tool",
        )
    )

    compiler = JSONCompiler()
    with open(new_sql) as out_file:
        out_file.write(compiler.visit(metadata))



import datetime

from omi import structure as s
from omi.oem_structures import oem_v15

_source_year = s.Field(
    name="year", description="Reference year", field_type="integer", unit=None
)

_target_year = s.Field(name="year", description=None, field_type=None, unit=None)

_target_resource = s.Resource(
    name="schema.table",
    schema=s.Schema(fields=[_target_year], foreign_keys=None, primary_key=None),
    dialect=None,
    encoding=None,
    path=None,
    profile=None,
    resource_format=None,
)

_target_year.resource = _target_resource


cc010 = s.License(
    name="Creative Commons Zero v1.0 Universal",
    identifier="CC0-1.0",
    path="https://creativecommons.org/publicdomain/zero/1.0/legalcode",
    other_references=None,
    text=None,
)

odbl10 = s.License(
    identifier="ODbL-1.0",
    name="Open Data Commons Open Database License 1.0",
    path="https://opendatacommons.org/licenses/odbl/1.0/",
    other_references=None,
    text=None,
)

odbl10_13 = s.License(
    name=None, identifier="ODbL-1.0", path=None, other_references=None, text=None
)

metadata_v_1_3 = s.OEPMetadata(
    name=None,
    title="Conceived Example Table Meant for Creating an Illustrative Metadata String thereof",
    identifier=None,
    description="An imaginary table that provides many features, offering a suitable source for metadata template entries",
    languages=["eng"],
    keywords=None,
    publication_date=None,
    context=None,
    spatial=s.Spatial(location=None, extent="Berlin", resolution="1 m"),
    temporal=s.Temporal(
        reference_date=datetime.datetime(2018, 11, 13),
        start=None,
        end=None,
        resolution=None,
        ts_orientation=None,
    ),
    sources=[
        s.Source(
            title="Technical review and evaluation of Issue",
            description="Study financed by Organisation describes Issue. The study is authored by Jon Doe and Erika Mustermann",
            path="https://doi.org/1.1/j.d.2000.01.001",
            licenses=[s.TermsOfUse(attribution="Publisher")],
        ),
        s.Source(
            title="Metastudy on Issue",
            description="Study financed by State Actor evaluates Issue in regions. The study is authored by Jane Doe and Otto Normal",
            path="https://doi.org/2.2/j.d.2022.02.022",
            licenses=[s.TermsOfUse(attribution="Publisher2")],
        ),
    ],
    terms_of_use=[
        s.TermsOfUse(
            lic=odbl10,
            instruction="You are free: To Share, To Create, To Adapt; As long as you: Attribute, Share-Alike, Keep open!",
            attribution="Institute",
        )
    ],
    contributions=[
        s.Contribution(
            contributor=s.Person(
                name="Person McHuman", email="person.mchuman@good-institute.net"
            ),
            date=datetime.datetime(2011, 1, 11, 0, 0, 0),
            obj=None,
            comment="Prepared the dataset",
        ),
        s.Contribution(
            contributor=s.Person(
                name="Indivia Mensch", email="indivia.mensch@gute-organisation.org"
            ),
            date=datetime.datetime(2012, 2, 12, 0, 0, 0),
            obj=None,
            comment="Fixed Metadata String and date format ",
        ),
    ],
    resources=[
        s.Resource(
            name="example.datatable",
            resource_format="PostgreSQL",
            schema=s.Schema(
                fields=[
                    s.Field(
                        name="id",
                        description="unambiguous unique numer",
                        unit=None,
                        field_type=None,
                    ),
                    s.Field(
                        name="component_id",
                        description="Identifying numer of component. May repeat due to several occurences of the same component.",
                        unit=None,
                        field_type=None,
                    ),
                    s.Field(
                        name="measurement",
                        description="Measured by Instrument",
                        unit="kWh",
                        field_type=None,
                    ),
                    s.Field(
                        name="reference",
                        description="Bibtex String that references the information source.",
                        unit=None,
                        field_type=None,
                    ),
                ],
                primary_key=None,
                foreign_keys=None,
            ),
            path=None,
            profile=None,
            encoding=None,
            dialect=None,
        )
    ],
    review=None,
    comment=None,
)

metadata_v_1_4 = s.OEPMetadata(
    name="oep_metadata_table_example_v14",
    title="Good example title",
    identifier="http://openenergyplatform.org/dataedit/view/model_draft/oep_metadata_table_example_v14",
    description="example metadata for example data",
    languages=["en-GB", "en-US", "de-DE", "fr-FR"],
    keywords=["example", "template", "test"],
    publication_date=datetime.datetime(2018, 6, 12, 0, 0),
    context=s.Context(
        homepage="https://reiner-lemoine-institut.de/szenariendb/",
        documentation="https://github.com/OpenEnergyPlatform/organisation/wiki/metadata",
        source_code="https://github.com/OpenEnergyPlatform/examples/tree/master/metadata",
        contact="https://github.com/Ludee",
        grant_number="03ET4057",
        funding_agency=s.Agency(
            name="Bundesministerium für Wirtschaft und Energie",
            logo="https://www.innovation-beratung-foerderung.de/INNO/Redaktion/DE/Bilder/Titelbilder/titel_foerderlogo_bmwi.jpg?__blob=poster&v=2",
        ),
        publisher=s.Agency(
            logo="https://reiner-lemoine-institut.de//wp-content/uploads/2015/09/rlilogo.png"
        ),
    ),
    spatial=s.Spatial(location=None, extent="europe", resolution="100 m"),
    temporal=s.Temporal(
        reference_date=datetime.datetime(2016, 1, 1, 0, 0),
        start=datetime.datetime(
            2017, 1, 1, 0, 0, tzinfo=datetime.timezone(datetime.timedelta(0, 3600))
        ),
        end=datetime.datetime(
            2017, 12, 31, 23, 0, tzinfo=datetime.timezone(datetime.timedelta(0, 3600))
        ),
        resolution="1 h",
        ts_orientation=s.TimestampOrientation.left,
        aggregation="sum",
    ),
    sources=[
        s.Source(
            title="OpenEnergyPlatform Metadata Example",
            description="Metadata description",
            path="https://github.com/OpenEnergyPlatform",
            licenses=[
                s.TermsOfUse(
                    lic=cc010,
                    instruction="You are free: To Share, To Create, To Adapt",
                    attribution="© Reiner Lemoine Institut",
                )
            ],
        ),
        s.Source(
            title="OpenStreetMap",
            description="A collaborative project to create a free editable map of the world",
            path="https://www.openstreetmap.org/",
            licenses=[
                s.TermsOfUse(
                    lic=odbl10,
                    instruction="You are free: To Share, To Create, To Adapt; As long as you: Attribute, Share-Alike, Keep open!",
                    attribution="© OpenStreetMap contributors",
                )
            ],
        ),
    ],
    terms_of_use=[
        s.TermsOfUse(
            lic=odbl10,
            instruction="You are free: To Share, To Create, To Adapt; As long as you: Attribute, Share-Alike, Keep open!",
            attribution="© Reiner Lemoine Institut © OpenStreetMap contributors",
        )
    ],
    contributions=[
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2016, 6, 16),
            obj="metadata",
            comment="Create metadata",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2016, 11, 22),
            obj="metadata",
            comment="Update metadata",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2016, 11, 22),
            obj="metadata",
            comment="Update header and license",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2017, 3, 16),
            obj="metadata",
            comment="Add license to source",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2017, 3, 28),
            obj="metadata",
            comment="Add copyright to source and license",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2017, 5, 30),
            obj="metadata",
            comment="Release metadata version 1.3",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2017, 6, 26),
            obj="metadata",
            comment="Move referenceDate into temporal and remove array",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2018, 7, 19),
            obj="metadata",
            comment="Start metadata version 1.4",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2018, 7, 26),
            obj="data",
            comment="Rename table and files",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2018, 10, 18),
            obj="metadata",
            comment="Add contribution object",
        ),
        s.Contribution(
            contributor=s.Person(name="christian-rli", email=None),
            date=datetime.datetime(2018, 10, 18),
            obj="metadata",
            comment="Add datapackage compatibility",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2018, 11, 2),
            obj="metadata",
            comment="Release metadata version 1.4",
        ),
        s.Contribution(
            contributor=s.Person(name="christian-rli", email=None),
            date=datetime.datetime(2019, 2, 5),
            obj="metadata",
            comment="Apply template structure to example",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2019, 3, 22),
            obj="metadata",
            comment="Hotfix foreignKeys",
        ),
        s.Contribution(
            contributor=s.Person(name="Ludee", email=None),
            date=datetime.datetime(2019, 7, 9),
            obj="metadata",
            comment="Release metadata version OEP-1.3.0",
        ),
    ],
    resources=[
        s.Resource(
            name="model_draft.oep_metadata_table_example_v14",
            path="http://openenergyplatform.org/dataedit/view/model_draft/oep_metadata_table_example_v14",
            profile="tabular-data-resource",
            resource_format="PostgreSQL",
            encoding="UTF-8",
            schema=s.Schema(
                fields=[
                    s.Field(
                        name="id",
                        description="Unique identifier",
                        field_type="serial",
                        unit=None,
                    ),
                    _source_year,
                    s.Field(
                        name="value",
                        description="Example value",
                        field_type="double precision",
                        unit="MW",
                    ),
                    s.Field(
                        name="geom",
                        description="Geometry",
                        field_type="geometry(Point, 4326)",
                        unit=None,
                    ),
                ],
                primary_key=["id"],
                foreign_keys=[
                    s.ForeignKey(
                        references=[
                            s.Reference(source=_source_year, target=_target_year)
                        ]
                    )
                ],
            ),
            dialect=s.Dialect(delimiter=None, decimal_separator="."),
        )
    ],
    review=s.Review(
        path="https://github.com/OpenEnergyPlatform/data-preprocessing/wiki",
        badge="platin",
    ),
    comment=s.MetaComment(
        metadata_info="Metadata documentation and explanation (https://github.com/OpenEnergyPlatform/organisation/wiki/metadata)",
        dates="Dates and time must follow the ISO8601 including time zone (YYYY-MM-DD or YYYY-MM-DDThh:mm:ss±hh)",
        units="Use a space between numbers and units (100 m)",
        languages="Languages must follow the IETF (BCP47) format (en-GB, en-US, de-DE)",
        licenses="License name must follow the SPDX License List (https://spdx.org/licenses/)",
        review="Following the OEP Data Review (https://github.com/OpenEnergyPlatform/data-preprocessing/wiki)",
        none="If not applicable use (null)",
    ),
)

############################################### oem v151 #########################################################
metadata_v_1_5 = oem_v15.OEPMetadata(
    name="oep_metadata_table_example_v151",
    title="Example title for metadata example - Version 1.5.1",
    identifier="http://openenergyplatform.org/dataedit/view/model_draft/oep_metadata_table_example_v151",
    description="This is an metadata example for example data. There is a corresponding table on the OEP for each metadata version.",
    languages=["en-GB", "en-US", "de-DE", "fr-FR"],
    subject=[
        oem_v15.Subject(
            name="energy",
            path="https://openenergy-platform.org/ontology/oeo/OEO_00000150",
        ),
        oem_v15.Subject(
            name="test dataset",
            path="https://openenergy-platform.org/ontology/oeo/OEO_00000408",
        ),
    ],
    keywords=["energy", "example", "template", "test"],
    publication_date=datetime.datetime(2022, 2, 15, 0, 0),
    context=oem_v15.Context(
        homepage="https://reiner-lemoine-institut.de/lod-geoss/",
        documentation="https://openenergy-platform.org/tutorials/jupyter/OEMetadata/",
        source_code="https://github.com/OpenEnergyPlatform/oemetadata/tree/master",
        contact="https://github.com/Ludee",
        grant_number="03EI1005",
        funding_agency=oem_v15.Agency(
            name="Bundesministerium für Wirtschaft und Klimaschutz",
            logo="https://commons.wikimedia.org/wiki/File:BMWi_Logo_2021.svg#/media/File:BMWi_Logo_2021.svg",
        ),
        publisher=oem_v15.Agency(
            logo="https://reiner-lemoine-institut.de//wp-content/uploads/2015/09/rlilogo.png"
        ),
    ),
    spatial=oem_v15.Spatial(location=None, extent="europe", resolution="100 m"),
    temporal=oem_v15.Temporal(
        reference_date=datetime.datetime(2016, 1, 1, 0, 0),
        timeseries_collection=[
            oem_v15.Timeseries(
                start=datetime.datetime(
                    2017,
                    1,
                    1,
                    0,
                    0,
                    tzinfo=datetime.timezone(datetime.timedelta(0, 3600)),
                ),
                end=datetime.datetime(
                    2017,
                    12,
                    31,
                    23,
                    0,
                    tzinfo=datetime.timezone(datetime.timedelta(0, 3600)),
                ),
                resolution="1 h",
                ts_orientation=oem_v15.TimestampOrientation.left,
                aggregation="sum",
            ),
            oem_v15.Timeseries(
                start=datetime.datetime(
                    2018,
                    1,
                    1,
                    0,
                    0,
                    tzinfo=datetime.timezone(datetime.timedelta(0, 3600)),
                ),
                end=datetime.datetime(
                    2019,
                    6,
                    1,
                    23,
                    0,
                    tzinfo=datetime.timezone(datetime.timedelta(0, 3600)),
                ),
                resolution="15 min",
                ts_orientation=oem_v15.TimestampOrientation.right,
                aggregation="sum",
            ),
        ],
    ),
    sources=[
        oem_v15.Source(
            title="OpenEnergyPlatform Metadata Example",
            description="Metadata description",
            path="https://github.com/OpenEnergyPlatform",
            licenses=[
                oem_v15.TermsOfUse(
                    lic=cc010,
                    instruction="You are free: To Share, To Create, To Adapt",
                    attribution="© Reiner Lemoine Institut",
                )
            ],
        ),
        oem_v15.Source(
            title="OpenStreetMap",
            description="A collaborative project to create a free editable map of the world",
            path="https://www.openstreetmap.org/",
            licenses=[
                oem_v15.TermsOfUse(
                    lic=odbl10,
                    instruction="You are free: To Share, To Create, To Adapt; As long as you: Attribute, Share-Alike, Keep open!",
                    attribution="© OpenStreetMap contributors",
                )
            ],
        ),
    ],
    terms_of_use=[
        oem_v15.TermsOfUse(
            lic=odbl10,
            instruction="You are free: To Share, To Create, To Adapt; As long as you: Attribute, Share-Alike, Keep open!",
            attribution="© Reiner Lemoine Institut © OpenStreetMap contributors",
        )
    ],
    contributions=[
        oem_v15.Contribution(
            contributor=oem_v15.Person(name="Ludee", email=None),
            date=datetime.datetime(2021, 11, 15),
            obj="metadata",
            comment="Release metadata version OEP-1.5.0",
        ),
        oem_v15.Contribution(
            contributor=oem_v15.Person(name="Ludee", email=None),
            date=datetime.datetime(2022, 2, 15),
            obj="metadata",
            comment="Release metadata version OEP-1.5.1",
        ),
    ],
    resources=[
        oem_v15.Resource(
            name="model_draft.oep_metadata_table_example_v151",
            path="http://openenergyplatform.org/dataedit/view/model_draft/oep_metadata_table_example_v151",
            profile="tabular-data-resource",
            resource_format="PostgreSQL",
            encoding="UTF-8",
            schema=oem_v15.Schema(
                fields=[
                    oem_v15.Field(
                        name="id",
                        description="Unique identifier",
                        field_type="serial",
                        unit=None,
                        is_about=[oem_v15.IsAbout(name=None, path=None)],
                        value_reference=[
                            oem_v15.ValueReference(value=None, name=None, path=None)
                        ],
                    ),
                    oem_v15.Field(
                        name="name",
                        description="Example name",
                        field_type="text",
                        unit=None,
                        is_about=[
                            oem_v15.IsAbout(
                                name="written name",
                                path="https://openenergy-platform.org/ontology/oeo/IAO_0000590",
                            )
                        ],
                        value_reference=[
                            oem_v15.ValueReference(value=None, name=None, path=None)
                        ],
                    ),
                    oem_v15.Field(
                        name="type",
                        description="Type of wind farm",
                        field_type="text",
                        unit=None,
                        is_about=[
                            oem_v15.IsAbout(
                                name="wind farm",
                                path="https://openenergy-platform.org/ontology/oeo/OEO_00000447",
                            )
                        ],
                        value_reference=[
                            oem_v15.ValueReference(
                                value="onshore",
                                name="onshore wind farm",
                                path="https://openenergy-platform.org/ontology/oeo/OEO_00000311",
                            ),
                            oem_v15.ValueReference(
                                value="offshore",
                                name="offshore wind farm",
                                path="https://openenergy-platform.org/ontology/oeo/OEO_00000308",
                            ),
                        ],
                    ),
                    oem_v15.Field(
                        name="year",
                        description="Example value",
                        field_type="double precision",
                        unit=None,
                        is_about=[
                            oem_v15.IsAbout(
                                name="year",
                                path="https://openenergy-platform.org/ontology/oeo/UO_0000036",
                            )
                        ],
                        value_reference=[
                            oem_v15.ValueReference(value=None, name=None, path=None)
                        ],
                    ),
                    oem_v15.Field(
                        name="value",
                        description="Example value",
                        field_type="double precision",
                        unit="MW",
                        is_about=[
                            oem_v15.IsAbout(
                                name="quantity value",
                                path="https://openenergy-platform.org/ontology/oeo/OEO_00000350",
                            )
                        ],
                        value_reference=[
                            oem_v15.ValueReference(value=None, name=None, path=None)
                        ],
                    ),
                    oem_v15.Field(
                        name="geom",
                        description="Geometry",
                        field_type="geometry(Point, 4326)",
                        unit=None,
                        is_about=[
                            oem_v15.IsAbout(
                                name="spatial region",
                                path="https://openenergy-platform.org/ontology/oeo/BFO_0000006",
                            )
                        ],
                        value_reference=[
                            oem_v15.ValueReference(
                                value="test", name="test", path="test"
                            )
                        ],
                    ),
                ],
                primary_key=["id"],
                foreign_keys=[
                    oem_v15.ForeignKey(
                        references=[
                            oem_v15.Reference(source=_source_year, target=_target_year)
                        ]
                    )
                ],
            ),
            dialect=oem_v15.Dialect(delimiter=None, decimal_separator="."),
        )
    ],
    databus_identifier="https://databus.dbpedia.org/kurzum/mastr/bnetza-mastr/01.04.00",
    databus_context="https://github.com/OpenEnergyPlatform/oemetadata/blob/master/metadata/latest/context.json",
    review=oem_v15.Review(
        path="https://github.com/OpenEnergyPlatform/data-preprocessing/issues",
        badge="Platinum",
    ),
    comment=oem_v15.MetaComment(
        metadata_info="Metadata documentation and explanation (https://github.com/OpenEnergyPlatform/oemetadata)",
        dates="Dates and time must follow the ISO8601 including time zone (YYYY-MM-DD or YYYY-MM-DDThh:mm:ss±hh)",
        units="Use a space between numbers and units (100 m)",
        languages="Languages must follow the IETF (BCP47) format (en-GB, en-US, de-DE)",
        licenses="License name must follow the SPDX License List (https://spdx.org/licenses/)",
        review="Following the OEP Data Review (https://github.com/OpenEnergyPlatform/data-preprocessing/blob/master/data-review/manual/review_manual.md)",
        null="If not applicable use: null",
        todo="If a value is not yet available, use: todo",
    ),
)


metadata_v_1_3_minimal = s.OEPMetadata()

metadata_v_1_4_minimal = s.OEPMetadata(identifier="id")

metadata_v_1_5_minimal = oem_v15.OEPMetadata(identifier="id")



from typing import Dict
from typing import Tuple

from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef
from rdflib.graph import Node
from rdflib.namespace import DCTERMS
from rdflib.namespace import FOAF
from rdflib.namespace import RDF
from rdflib.namespace import XSD

from omi import structure
from omi.dialects.base.compiler import Compiler
from omi.dialects.rdf.licenses import LICENSE_DICT
from omi.dialects.rdf.namespace import ADMS
from omi.dialects.rdf.namespace import DCAT
from omi.dialects.rdf.namespace import DCATDE
from omi.dialects.rdf.namespace import OEO
from omi.dialects.rdf.namespace import SCHEMA
from omi.dialects.rdf.namespace import SKOS
from omi.dialects.rdf.namespace import SPDX

LANG_DICT = {
    "eng": "http://publications.europa.eu/resource/authority/language/ENG",
    "en-GB": "http://publications.europa.eu/resource/authority/language/ENG",
    "ger": "http://publications.europa.eu/resource/authority/language/GER",
    "en-US": "",
    "de-DE": "http://publications.europa.eu/resource/authority/language/GER",
    "fr-FR": "",
}


class RDFCompiler(Compiler):
    def __init__(self, graph: Graph = None):
        if graph is None:
            self.graph = Graph()
            self.graph.namespace_manager.bind("foaf", FOAF)
            self.graph.namespace_manager.bind("dct", DCTERMS)
            self.graph.namespace_manager.bind("dcat", DCAT)
            self.graph.namespace_manager.bind("dcatde", DCATDE)
            self.graph.namespace_manager.bind("oeo", OEO)
            self.graph.namespace_manager.bind("schema", SCHEMA)
            self.graph.namespace_manager.bind("skos", SKOS)
            self.graph.namespace_manager.bind("adms", ADMS)
            self.graph.namespace_manager.bind("spdx", SPDX)
        else:
            self.graph = graph

    def visit_agency(self, agency: structure.Agency, *args, **kwargs):
        node = BNode()
        self._add_literal_or_None(node, DCTERMS.title, agency.name)
        self._add_literal_or_None(node, OEO.has_logo, agency.logo)
        return node

    def visit_context(self, context: structure.Context, *args, **kwargs):

        return (
            Literal(context.homepage),
            Literal(context.contact),
            Literal(context.documentation),
            Literal(context.source_code),
            Literal(context.grant_number),
            self.visit(context.funding_agency),
            self.visit(context.publisher),
        )

    def visit_person(self, person: structure.Person, *args, **kwargs):

        node = BNode()
        self.graph.add((node, RDF.type, FOAF.Person))
        self._add_literal_or_None(node, FOAF.name, person.name)
        self._add_literal_or_None(node, FOAF.mbox, person.email)
        return node

    def visit_contribution(self, contribution: structure.Contribution, *args, **kwargs):
        c = BNode()
        self.graph.add(
            (
                c,
                DCTERMS.contributor,
                self.visit(contribution.contributor, *args, **kwargs),
            )
        )
        self._add_literal_or_None(
            c, OEO.date, contribution.date.strftime("%Y-%m-%d"), datatype=XSD.date
        )
        self._add_literal_or_None(c, OEO.comment, contribution.comment)
        self._add_literal_or_None(c, OEO.object, contribution.object)
        return c

    def visit_language(self, language: structure.Language, *args, **kwargs):
        return URIRef(LANG_DICT[language])

    def visit_spatial(self, spatial: structure.Spatial, *args, **kwargs):
        node = BNode()
        self._add_literal_or_None(node, SKOS.prefLabel, spatial.extent)
        self._add_literal_or_None(node, OEO.has_spatial_resolution, spatial.resolution)
        self._add_literal_or_None(node, OEO.location, spatial.location)
        return node

    def visit_temporal(self, temporal: structure.Temporal, *args, **kwargs):
        node = BNode()
        self.graph.add((node, RDF.type, DCTERMS.PeriodOfTime))
        self._add_literal_or_None(node, SCHEMA.startDate, temporal.ts_start)
        self._add_literal_or_None(node, SCHEMA.endDate, temporal.ts_end)
        self._add_literal_or_None(node, OEO.has_time_resolution, temporal.ts_resolution)
        self._add_literal_or_None(node, OEO.referenceDate, temporal.reference_date)
        self.graph.add(
            (node, OEO.has_timestamp_alignment, self.visit(temporal.ts_orientation))
        )
        self._add_literal_or_None(node, OEO.uses_aggregation, temporal.aggregation)
        return node

    def visit_timestamp_orientation(
        self, ts_orientation: structure.TimestampOrientation, *args, **kwargs
    ):
        if ts_orientation == structure.TimestampOrientation.left:
            return OEO.left_orientation
        elif ts_orientation == structure.TimestampOrientation.middle:
            return OEO.middle_orientation
        elif ts_orientation == structure.TimestampOrientation.right:
            return OEO.right_orientation
        else:
            raise Exception("Unknown timestamp orientation")

    def visit_source(self, source: structure.Source, *args, **kwargs):
        node = BNode()
        self._add_literal_or_None(node, DCTERMS.title, source.title)
        self._add_literal_or_None(node, DCTERMS.description, source.description)
        self._add_literal_or_None(node, FOAF.page, source.path)
        for l in source.licenses:
            li = self.visit(l, **kwargs)
            self.graph.add((node, OEO.has_terms_of_use, li))
        return node

    def visit_terms_of_use(self, tou: structure.TermsOfUse, *args, **kwargs):
        node = BNode()
        self._add_literal_or_None(
            node, DCATDE.licenseAttributionByText, tou.attribution
        )
        self._add_literal_or_None(node, OEO.has_instruction, tou.instruction)
        self.graph.add((node, DCAT.license, self.visit(tou.license, **kwargs)))
        return node

    def visit_license(self, lic: structure.License, *args, license_dict=None, **kwargs):
        if False:  # lic.name in LICENSE_DICT:
            li = URIRef(LICENSE_DICT[lic.identifier])
        else:
            if license_dict is not None and lic.identifier in license_dict:
                li = license_dict[lic.identifier]
            else:
                li = BNode()
                self.graph.add((li, RDF.type, DCAT.LicenseDocument))
                if lic.other_references:
                    for ref in lic.other_references:
                        self._add_literal_or_None(li, SPDX.seeAlso, ref)
                self._add_literal_or_None(li, FOAF.page, lic.path)
                self._add_literal_or_None(li, SPDX.licenseId, lic.identifier)
                if lic.text:
                    self._add_literal_or_None(li, SPDX.licenseText, lic.text)
                self._add_literal_or_None(li, SPDX.name, lic.name)
                if license_dict is not None and lic.identifier is not None:
                    license_dict[lic.identifier] = li
        return li

    def visit_resource(
        self,
        resource: structure.Resource,
        *args,
        resource_dict: Dict[str, Tuple[Node, Dict[str, Node]]] = None,
        **kwargs
    ):
        s = BNode()
        self.graph.add((s, RDF.type, DCAT.Distribution))
        self._add_literal_or_None(s, DCTERMS.title, resource.name)
        self._add_literal_or_None(s, DCAT.accessURL, resource.path)
        self._add_literal_or_None(s, OEO.has_format, resource.format)  # dct:format ?
        self._add_literal_or_None(s, OEO.profile, resource.profile)
        self._add_literal_or_None(s, OEO.encoding, resource.encoding)
        if resource.dialect is not None:
            self.graph.add((s, OEO.has_dialect, self.visit(resource.dialect)))
        fields, pks, fks = self.visit(resource.schema, s, resource_dict=resource_dict)
        for f in fields:
            self.graph.add((s, OEO.field, f))
        for pk in pks:
            self.graph.add((s, OEO.primaryKey, pk))
        for fk in fks:
            self.graph.add((s, OEO.has_foreignKey, fk))
        return s

    def visit_schema(
        self,
        schema: structure.Schema,
        *args,
        resource_dict: Dict[str, Tuple[Node, Dict[str, Node]]] = None,
        **kwargs
    ):
        resource_node = args[0]
        field_dict = dict()
        fields = []
        resource_dict = resource_dict or dict()
        for field in schema.fields:
            fnode = self.visit(field, *args, **kwargs)
            fields.append(fnode)
            field_dict[field.name] = fnode

        pks = [field_dict[pk] for pk in schema.primary_key or []]

        resource_name = schema.fields[0].resource.name

        if resource_name in resource_dict:
            resource_dict[resource_name][1].update(field_dict)
        else:
            resource_dict[resource_name] = (resource_node, field_dict)

        fks = [
            self.visit(fk, *args, resource_dict=resource_dict, **kwargs)
            for fk in schema.foreign_keys or []
        ]

        return fields, pks, fks

    def visit_dialect(self, dialect: structure.Dialect, *args, **kwargs):
        node = BNode()
        self._add_literal_or_None(node, OEO.delimiter, dialect.delimiter)
        self._add_literal_or_None(node, OEO.decimalSeparator, dialect.decimal_separator)
        return node

    def visit_field(self, field: structure.Field, *args, **kwargs):
        field_uri = BNode()
        self.graph.add((field_uri, RDF.type, OEO.DatabaseField))
        self._add_literal_or_None(field_uri, DCTERMS.title, field.name)
        self._add_literal_or_None(field_uri, DCTERMS.description, field.description)
        self._add_literal_or_None(field_uri, OEO.type, field.type)
        self._add_literal_or_None(field_uri, OEO.unit, field.unit)
        return field_uri

    def visit_foreign_key(
        self, foreign_key: structure.ForeignKey, *args, resource_dict=None, **kwargs
    ):
        fk_node = BNode()
        for r in foreign_key.references:
            self.graph.add(
                (
                    fk_node,
                    OEO.has_reference,
                    self.visit(r, *args, resource_dict=resource_dict, **kwargs),
                )
            )
        return fk_node

    def visit_reference(
        self,
        reference: structure.Reference,
        *args,
        resource_dict: Dict[str, Tuple[Node, Dict[str, Node]]] = None,
        **kwargs
    ):
        r_node = BNode()
        self.graph.add(
            (r_node, OEO.has_source, self._get_field(reference.source, resource_dict))
        )
        target = self._get_or_create_field(
            reference.target, resource_dict, *args, **kwargs
        )
        self.graph.add((r_node, OEO.has_target, target))
        return r_node

    def visit_review(self, review: structure.Review, *args, **kwargs):
        node = BNode()
        self.graph.add((node, FOAF.page, URIRef(review.path)))
        self._add_literal_or_None(node, OEO.has_badge, review.badge)
        return node

    def visit_meta_comment(self, comment: structure.MetaComment, *args, **kwargs):
        com = BNode()
        self._add_literal_or_None(com, OEO.metadata_info, comment.metadata_info)
        self._add_literal_or_None(com, OEO.dates_info, comment.dates)
        self._add_literal_or_None(com, OEO.units_info, comment.units)
        self._add_literal_or_None(com, OEO.languages_info, comment.languages)
        self._add_literal_or_None(com, OEO.licenses_info, comment.licenses)
        self._add_literal_or_None(com, OEO.review_info, comment.review)
        self._add_literal_or_None(com, OEO.none_info, comment.none)
        return com

    def visit_metadata(self, metadata: structure.OEPMetadata, *args, **kwargs):
        datasetURI = URIRef(metadata.identifier)

        license_dict = kwargs.get("license_dict", {})

        self.graph.add((datasetURI, RDF.type, DCAT.Dataset))
        self._add_literal_or_None(datasetURI, ADMS.Identifier, metadata.name)
        self._add_literal_or_None(datasetURI, DCTERMS.title, metadata.title)
        self._add_literal_or_None(datasetURI, DCTERMS.description, metadata.description)

        for lang in metadata.languages:
            self._add_literal_or_None(datasetURI, DCTERMS.language, self.visit(lang))

        for k in metadata.keywords:
            self._add_literal_or_None(datasetURI, DCAT.keyword, k)

        self.graph.add(
            (
                datasetURI,
                OEO.publicationDate,
                Literal(
                    metadata.publication_date.strftime("%Y-%m-%d"), datatype=XSD.date
                ),
            )
        )

        homepage, contact, documentation, source_code, grant_number, funding_agency, publisher = self.visit(
            metadata.context
        )

        self.graph.add((datasetURI, FOAF.homepage, homepage))
        self.graph.add((datasetURI, DCAT.contactpoint, contact))
        self.graph.add((datasetURI, OEO.documentation, documentation))
        self.graph.add((datasetURI, OEO.sourceCode, source_code))
        self.graph.add((datasetURI, OEO.grantNo, grant_number))

        if funding_agency:
            self.graph.add((datasetURI, OEO.has_funding_agency, funding_agency))

        if publisher:
            self.graph.add((datasetURI, OEO.has_publisher, publisher))

        self.graph.add((datasetURI, DCTERMS.spatial, self.visit(metadata.spatial)))
        self.graph.add((datasetURI, DCTERMS.temporal, self.visit(metadata.temporal)))
        for s in metadata.sources:
            self.graph.add(
                (datasetURI, DCTERMS.source, self.visit(s, license_dict=license_dict))
            )
        for l in metadata.license:
            self.graph.add(
                (
                    datasetURI,
                    OEO.has_terms_of_use,
                    self.visit(l, license_dict=license_dict),
                )
            )

        for c in metadata.contributions:
            self.graph.add((datasetURI, OEO.has_contribution, self.visit(c)))
        for r in metadata.resources:
            self.graph.add((datasetURI, OEO.has_resource, self.visit(r)))

        self.graph.add((datasetURI, OEO.has_review, self.visit(metadata.review)))

        self.graph.add(
            (
                datasetURI,
                OEO.metadataLicense,
                URIRef("https://creativecommons.org/publicdomain/zero/1.0/legalcode"),
            )
        )

        self.graph.add((datasetURI, OEO.comment, self.visit(metadata.comment)))

        return datasetURI

    def _get_or_create_field(
        self,
        field: structure.Field,
        resource_dict: Dict[str, Tuple[Node, Dict[str, Node]]],
        *args,
        **kwargs
    ):
        try:
            if field.resource.name not in resource_dict:
                self.visit_resource(field.resource, resource_dict=resource_dict)
            return self._get_field(field, resource_dict)
        except FieldNotFoundError:
            f = self.visit_field(field, *args, **kwargs)
            res_node, field_dict = resource_dict[field.resource.name]
            self.graph.add((res_node, OEO.field, f))
            field_dict[field.name] = f
            return f

    def _get_field(
        self,
        field: structure.Field,
        resource_dict: Dict[str, Tuple[Node, Dict[str, Node]]],
    ):
        res = resource_dict.get(field.resource.name)
        if res is None:
            raise ResourceNotFoundError(field.resource.name)
        node = res[1].get(field.name)
        if node is None:
            raise FieldNotFoundError(field.name)
        return node

    def _add_literal_or_None(self, subject: Node, predicate, obj: str, **kwargs):
        if obj is not None:
            self.graph.add((subject, predicate, Literal(obj, **kwargs)))


class FieldNotFoundError(Exception):
    pass


class ResourceNotFoundError(Exception):
    pass



import json

from omi.dialects.oep.compiler import JSONCompiler, JSONCompilerOEM15

from ..internal_structures import metadata_v_1_4, metadata_v_1_5


def test_compiler_v1_5():
    compiler = JSONCompilerOEM15()
    with open("tests/data/metadata_v15.json", "r", encoding="utf-8") as _input_file:
        expected_result = json.load(_input_file)
        result = compiler.visit(metadata_v_1_5)
        assert_equal(expected_result, result)


def test_compiler_v1_4():
    compiler = JSONCompiler()
    with open("tests/data/metadata_v14.json", "r", encoding="utf-8") as _input_file:
        expected_result = json.load(_input_file)
        result = compiler.visit(metadata_v_1_4)
        assert_equal(expected_result, result)


def assert_equal(first, second, nulls=frozenset(["none"]), **kwargs):
    if isinstance(first, dict):
        assert_dict_equal(first, second, **kwargs)
    elif isinstance(first, list):
        assert_list_equal(first, second, **kwargs)
    else:
        if first in nulls and second is None:
            return
        assert first == second


def assert_dict_equal(first, second, **kwargs):
    assert_list_equal(list(first.keys()), list(second.keys()), disregard_ordering=True)
    for key in first.keys():
        if key not in second:
            raise Exception('Key "{}" missing in {}'.format(key, second))
        assert_equal(first[key], second[key], **kwargs)


def assert_list_equal(first, second, disregard_ordering=False, **kwargs):
    assert len(first) == len(second), "Length mismatch ({}!={}) for {} and {}".format(
        len(first), len(second), first, second
    )

    if disregard_ordering:
        remaining = list(second)
        for expected in first:
            to_remove = None
            for i, result in enumerate(remaining):
                try:
                    assert_equal(
                        expected,
                        result,
                        disregard_ordering=disregard_ordering,
                        **kwargs
                    )
                except:
                    pass
                else:
                    to_remove = i
            if to_remove is not None:
                del remaining[to_remove]
            else:
                raise AssertionError(
                    "Element not found: {} in {}".format(expected, second)
                )
        if remaining:
            raise AssertionError(
                "Elements not found: {} in {}".format(remaining, first)
            )
    else:
        for i, (l, r) in enumerate(zip(first, second)):
            try:
                assert_equal(l, r, **kwargs)
            except:
                print("Mismatch in element {}: {}!={}".format(i, l, r))
                raise



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -momi` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``omi.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``omi.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import click

from omi.dialects import get_dialect


@click.group()
def grp():
    pass


@grp.command("translate")
@click.option("-f", help="Dialect identifier of the input")
@click.option("-t", default="oep-v1.5", help="Dialect identifier to translate to")
@click.option("-o", default=None, help="Output file")
@click.argument("file_path")
def translate(f, t, o, file_path):
    with open(file_path, "r", encoding="utf-8") as infile:
        from_dialect = get_dialect(f)()
        obj = from_dialect.parse(infile.read())
        to_dialect = get_dialect(t)()
        s = to_dialect.compile_and_render(obj)
        if o:
            with open(o, "w", encoding="utf-8") as outfile:
                outfile.write(s)
        else:
            print(s)


cli = click.CommandCollection(sources=[grp])


def main():
    cli()



from datetime import datetime
from enum import Enum
from typing import Iterable


class Compilable:
    """
    An abstract class for all metadata components.
    """

    __compiler_name__ = None
    """Used to identify the appropriate compiler function for this structure"""

    __required__ = None
    __optional__ = None

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ",".join("{}={}".format(key, val) for key, val in self.__dict__.items()),
        )

    def __lt__(self, other):
        for key in sorted(self.__dict__):
            s = getattr(self, key)
            o = getattr(other, key)
            if s is None:
                return True
            elif s < o:
                return True
            elif s > o:
                return False
        return False

    def get_missing_fields(self):
        for key in sorted(self.__dict__):
            if key in self.__required__:
                if s is None:
                    yield key
                v = getattr(self, key)
                if isinstance(v, Compilable):
                    for x in v.get_missing_fields():
                        yield key + "." + x


class Language(Compilable):
    __compiler_name__ = "language"


class Spatial(Compilable):
    __compiler_name__ = "spatial"

    def __init__(
        self, location: str = None, extent: str = None, resolution: str = None
    ):
        self.location = location
        self.extent = extent
        self.resolution = resolution


class TimestampOrientation(Compilable, Enum):
    __compiler_name__ = "timestamp_orientation"
    left = 0
    middle = 1
    right = 2

    @staticmethod
    def create(value: str) -> "TimestampOrientation":
        if value == "left":
            return TimestampOrientation.left
        elif value == "middle":
            return TimestampOrientation.middle
        elif value == "right":
            return TimestampOrientation.right
        else:
            raise Exception("Unknown timestamp orientation:", value)


class Temporal(Compilable):
    __compiler_name__ = "temporal"

    def __init__(
        self,
        reference_date: datetime = None,
        start: datetime = None,
        end: datetime = None,
        resolution: str = None,
        ts_orientation: TimestampOrientation = None,
        aggregation: str = None,
    ):  # TODO: This should not be a string... maybe
        # we should use datetime instead?
        self.reference_date = reference_date
        self.ts_start = start
        self.ts_end = end
        self.ts_resolution = resolution
        self.ts_orientation = ts_orientation
        self.aggregation = aggregation


class License(Compilable):
    __compiler_name__ = "license"

    def __init__(
        self,
        name: str = None,
        identifier: str = None,
        text: str = None,
        path: str = None,
        other_references: Iterable[str] = None,
        comment: str = None,
    ):
        self.name = name
        self.path = path
        self.identifier = identifier
        self.other_references = other_references
        self.text = text
        self.comment = comment

    @staticmethod
    def instance_name_from_id(identifier: str):
        return "L_" + identifier.replace("-", "_").replace(".", "_").replace(
            "+", "_plus"
        )


class TermsOfUse(Compilable):
    __compiler_name__ = "terms_of_use"

    def __init__(
        self, instruction: str = None, attribution: str = None, lic: License = None
    ):
        self.instruction = instruction
        self.attribution = attribution
        self.license = lic


class Source(Compilable):
    __compiler_name__ = "source"

    def __init__(
        self,
        title: str = None,
        description: str = None,
        path: str = None,
        licenses: Iterable[TermsOfUse] = None,
    ):
        self.title = title
        self.description = description
        self.path = path
        self.licenses = licenses


class Person(Compilable):
    __compiler_name__ = "person"

    def __init__(self, name: str = None, email: str = None):
        self.name = name
        self.email = email


class Contribution(Compilable):
    __compiler_name__ = "contribution"

    def __init__(
        self,
        contributor: Person = None,
        date: datetime = None,
        obj: str = None,
        comment: str = None,
    ):
        self.contributor = contributor
        self.date = date
        self.object = obj
        self.comment = comment


class Field(Compilable):
    __compiler_name__ = "field"

    def __init__(
        self,
        name: str = None,
        description: str = None,
        field_type: str = None,
        unit: str = None,
        resource: "Resource" = None,
    ):
        self.name = name
        self.description = description
        self.type = field_type
        self.unit = unit
        self.resource = resource

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ",".join(
                "{}={}".format(key, val)
                for key, val in self.__dict__.items()
                if key != "resource"
            ),
        )


class Agency(Compilable):
    __compiler_name__ = "agency"

    def __init__(self, name: str = None, logo: str = None):
        self.name = name
        self.logo = logo


class Context(Compilable):
    __compiler_name__ = "context"

    def __init__(
        self,
        homepage: str = None,
        documentation: str = None,
        source_code: str = None,
        contact: str = None,
        grant_number: str = None,
        funding_agency: Agency = None,
        publisher: Agency = None,
    ):
        self.homepage = homepage
        self.documentation = documentation
        self.source_code = source_code
        self.contact = contact
        self.grant_number = grant_number
        self.funding_agency = funding_agency
        self.publisher = publisher


class Reference(Compilable):
    __compiler_name__ = "reference"

    def __init__(self, source: Field = None, target: Field = None):
        self.source = source
        self.target = target


class ForeignKey(Compilable):
    __compiler_name__ = "foreign_key"

    def __init__(self, references: Iterable[Reference] = None):
        self.references = references


class Schema(Compilable):
    __compiler_name__ = "schema"

    def __init__(
        self,
        fields: Iterable[Field] = None,
        primary_key: Iterable[str] = None,
        foreign_keys: Iterable[ForeignKey] = None,
    ):
        self.fields = fields
        self.primary_key = primary_key
        self.foreign_keys = foreign_keys


class Dialect(Compilable):
    __compiler_name__ = "dialect"

    def __init__(self, delimiter: str = None, decimal_separator: str = None):
        self.delimiter = delimiter
        self.decimal_separator = decimal_separator


class Resource(Compilable):
    __compiler_name__ = "resource"

    def __init__(
        self,
        name: str = None,
        path: str = None,
        profile: str = None,
        resource_format: str = None,
        encoding: str = None,
        schema: Schema = None,
        dialect: Dialect = None,
    ):
        self.name = name
        self.path = path
        self.profile = profile
        self.format = resource_format
        self.encoding = encoding
        self.schema = schema
        if schema is not None:
            for field in schema.fields:
                field.resource = self
        self.dialect = dialect


class MetaComment(Compilable):
    __compiler_name__ = "meta_comment"

    def __init__(
        self,
        metadata_info: str = None,
        dates: str = None,
        units: str = None,
        languages: str = None,
        licenses: str = None,
        review: str = None,
        none: str = None,
    ):
        self.metadata_info = metadata_info
        self.dates = dates
        self.units = units
        self.languages = languages
        self.licenses = licenses
        self.review = review
        self.none = none


class Review(Compilable):
    __compiler_name__ = "review"

    def __init__(self, path: str = None, badge: str = None):
        self.path = path
        self.badge = badge


class OEPMetadata(Compilable):
    __compiler_name__ = "metadata"
    __required__ = ["id"]

    def __init__(
        self,
        name: str = None,
        title: str = None,
        identifier: str = None,
        description: str = None,
        languages: Iterable[Language] = None,
        keywords: Iterable[str] = None,
        publication_date: datetime = None,
        context: Context = None,
        spatial: Spatial = None,
        temporal: Temporal = None,
        sources: Iterable[Source] = None,
        terms_of_use: Iterable[TermsOfUse] = None,
        contributions: Iterable[Contribution] = None,
        resources: Iterable[Resource] = None,
        review: Review = None,
        comment: MetaComment = None,
    ):
        self.name = name
        self.title = title
        self.identifier = identifier
        self.description = description
        self.languages = languages
        self.keywords = keywords
        self.publication_date = publication_date
        self.context = context
        self.spatial = spatial
        self.temporal = temporal
        self.sources = sources
        self.license = terms_of_use
        self.contributions = contributions
        self.resources = resources
        self.review = review
        self.comment = comment

    def has_keywords(self):
        return self.keywords is not None


from rdflib.graph import Graph

from omi.dialects.base.renderer import Renderer


class GraphRenderer(Renderer):
    def render(self, inp: Graph, *args, **kwargs):
        return inp.serialize(format="ttl")
